import os
import numpy as np
import sciris as sc
import starsim as ss
import pylab as pl

reps = 10_000
keys = {'0.5.2':'modulo', '0.5.3':'roulette'}
filenames = {k:f'rng_stat_{k}.json' for k in keys.values()}


def make_sim(seed):
    pars = dict(
        rand_seed = seed,
        n_agents = 1000,
        end = 2020,
        diseases = ['sir', 'sis'],
        networks = 'random',
        verbose = 0,
    )
    sim = ss.Sim(pars)
    return sim

def run_sim_summary(seed):
    sim = make_sim(seed)
    sim.run()
    return sim.summary

def plot_sim():
    sim = make_sim()
    fig = sim.run().plot()
    return fig

def run_sims():
    ver = ss.__version__
    sc.heading(f'Running {ver}...')
    T = sc.timer()
    reslist = sc.parallelize(run_sim_summary, np.arange(reps))
    res = dict()
    for k in reslist[0].keys():
        res[k] = []
    for entry in reslist:
        for k in res.keys():
            res[k].append(entry[k])
    key = keys[ver]
    filename = filenames[key]
    sc.savejson(filename, res)
    T.toc()
    return

def get_raw():
    raw = sc.objdict()
    for k,filename in filenames.items():
        if not os.path.exists(filename):
            run_sims()
            print('Now change branch (if needed) and re-run')
            break
        raw[k] = sc.loadjson(filename)
    return raw

def get_res(raw):
    res = sc.objdict()
    for k,v in raw.items():
        res[k] = sc.objdict()
        for k2,v2 in v.items():
            mean = np.mean(v2)
            sem = np.std(v2)/np.sqrt(len(v2))
            res[k][k2] = sc.objdict(mean=mean, sem=sem)
    return res

def plot(raw, res):
    keys = res[0].keys()
    nkeys = len(keys)
    
    sc.options(dpi=150)
    fig = pl.figure(figsize=(16,16))
    count = 0
    for i,k in enumerate(keys):
        mod = res.modulo[k]
        rou = res.roulette[k]
        z = 2
        modmin = mod.mean - z*mod.sem
        modmax = mod.mean + z*mod.sem
        roumin = rou.mean - z*rou.sem
        roumax = rou.mean + z*rou.sem
        statsig = True if (modmax < roumin) or (modmin > roumax) else False
        for i2,k2 in enumerate(res.keys()):
            count += 1
            this = res[k2][k]
            title = f'{k2}\n{k}\n{this.mean:n}±{2*this.sem:n}'
            if statsig:
                title += '\n(mismatch!)'
            pl.subplot(nkeys//2,4,count)
            pl.hist(raw[k2][k], bins=int(np.sqrt(reps)))
            pl.title(title)
        
    sc.figlayout()
    pl.show()
    return fig
    
    
if __name__ == '__main__':
    raw = get_raw()
    res = get_res(raw)
    fig = plot(raw, res)


"""
What -- this shows no difference in the number of infections!!
"""

import numpy as np
import pylab as pl
import sciris as sc

n = 100
reps = 10_000
nc = 4
thresh = 1-(0.5/n)
sc.options(dpi=200)
pl.figure(figsize=(10,8))

d = sc.objdict(defaultdict=sc.autolist)

def matsort(arr):
    s0 = arr.sum(axis=0)
    s1 = arr.sum(axis=1)
    o0 = np.argsort(s0)
    o1 = np.argsort(s1)
    arr = arr[o1,:][:,o0]
    return arr

for rep in range(reps):

    # Truly random
    A = np.random.rand(n,n)
    A = matsort(A)
    
    # Modulo
    b1 = np.random.rand(n,1)
    b2 = np.random.rand(1,n)
    B1 = np.tile(b1, (1,n))
    B2 = np.tile(b2, (n,1))
    B = (B1 + B2) % 1.0
    B = matsort(B)
    
    # Thresholded
    At = A>thresh
    Bt = B>thresh
    v = sc.objdict()
    v.Asum = A.sum()
    v.Bsum = B.sum()
    v.Atsum = (At.sum(axis=1) > 0).sum()
    v.Btsum = (Bt.sum(axis=1) > 0).sum()
    
    for k,val in v.items():
        d[k] += val
    
    
    
#%% Plotting
labels = ['Truly random', 'Modulo']
reslabel = dict(
    Asum='Truly random, number of edge transmissions',
    Bsum='Modulo, number of edge transmissions',
    Atsum='Truly random, number of infections',
    Btsum='Modulo number of infections',
)
for i,k,val in d.enumitems():
    pl.subplot(2,2,i+1)
    pl.hist(val, bins=int(np.sqrt(reps)))
    pl.title(f'{reslabel[k]} {np.array(val).mean():n}')

sc.figlayout()
pl.show()



"""
Vignette 09: Custom networks
TO RESOLVE:
    - what should the default networks be?
    - how should network participation be specified? e.g. not everyone will participate in the MSM or FSW networks
    - what should happen if network properties depend on disease attributes? (e.g. Romesh's example)
    - what should happen if the network depends on custom people attributes?
"""

import stisim as ss
import numpy as np


##################
# @RomeshA
##################
people = ss.People.from_households(n=1000, <some household structure inputs>)  # Creates a household layer
people = ss.People(n=1000) # No contact layers are present
people.contacts['msm'] = ss.SexualMixingLayer(label='msm', people=people, eligible='m',mean_partners=2,dur=scipy.stats.lognorm(loc=5, s=3))
people.contacts['msm'] = ss.SexualMixingLayer(label='msm', people=people, eligible='m',mean_partners=2,dur=ss.lognormal(mean=5, std=3))

# What if the layer depends on some of the disease attributes?
# e.g.

class CasualMixingLayer(ss.SexualMixingLayer):
    def update(self, sim):
        n_contacts = self.mean_partners*np.ones(sim.n)
        n_contacts[sim.people.gonorrhea.symptomatic] *= 0.5
        super().update(sim, n_contacts)

people = ss.People(n=1000) # No contact layers are present
sim = ss.Sim(modules=[ss.HPV(), ss.Gonorrhea()], people=people)
sim.initialize_modules()
people.contacts['msm'] = ss.CasualMixingLayer(label='casual', sim=sim, eligible='m',mean_partners=2,dur=scipy.stats.lognorm(loc=5, s=3))


##################
# @robynstuart
##################
# Case 1: in-built network options
# Some defaults for networks can be selected via 'network_structure' parameter (akin to 'pop_type' in covasim):
# The default, which is an alias for sim = ss.Sim(networks='mf', modules=...) & equivalent to sim = ss.Sim(modules=...)
sim = ss.Sim(network_structure='random', modules=ss.gonorrhea())
# The 'hybrid' option adds MSM, & is an alias for adding an MSM state and adding the network
sim = ss.Sim(network_structure='hybrid', modules=ss.gonorrhea())
# This is an alias for:
states = ss.StochState('msm', bool, distdict=dict(dist='choice', par1=[1, 0], par2=[0.1, 0.9]))
sim = ss.Sim(states=states, networks=['mf', 'msm'], modules=ss.gonorrhea())

# Each disease should have default beta layer values for each built-in network
# The module should use the beta value for whichever networks have been added, e.g.
syph = ss.syphylis(
    beta=0.0008,  # Overall baseline beta. By convention this should represent M->F transmission risk
    beta_layer=dict(mf=[1, 0.5], msm=[10, 10])
)
sim = ss.Sim(network_structure='hybrid', modules=syph)


# Case 2: adding a custom network
ppl = ss.People(100)
nwk = ss.mf()
syph = ss.syphilis()
sim['syphilis']['beta']['mf'] *= 2  # Increase beta for syph transmission
sim = ss.Sim(people=ppl, networks=nwk, modules=syph)


# Case 3: adding a custom network that needs specific people attributes
is_fsw = ss.StochState('fsw', bool, distdict=dict(dist='choice', par1=[1, 0], par2=[0.05, 0.95]))
is_client = ss.StochState('client', bool, distdict=dict(dist='choice', par1=[1, 0], par2=[0.15, 0.85]))
ppl = ss.People(100, states=[is_fsw, is_client])
sex_work = sex_work()  # Question: what constraints are there on this? Must it be a ss.Network or could it be anything?
syph = ss.syphilis()
sim = ss.Sim(people=ppl, networks=['mf', sex_work], modules=syph)






# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.

import os
import sys
import sciris as sc
import starsim as ss

# Set environment
os.environ['SPHINX_BUILD'] = 'True' # This is used so ss.options.set('jupyter') doesn't reset the Matplotlib renderer
on_rtd = os.environ.get('READTHEDOCS') == 'True'


# -- Project information -----------------------------------------------------

project = 'Starsim'
copyright = f'2023 - {sc.now().year}, Bill & Melinda Gates Foundation. All rights reserved. These docs were built for Starsim v{ss.__version__}'
author = 'Starsim Collective'

# The short X.Y version
version = ss.__version__

# The full version, including alpha/beta/rc tags
release = ss.__version__


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here
extensions = [
    "sphinx.ext.autodoc",  # Core Sphinx library for auto html doc generation from docstrings
    "sphinx.ext.autosummary",  # Create neat summary tables for modules/classes/methods etc -- causes warnings with Napoleon however
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",  # Add a link to the Python source code for classes, functions etc.
    "sphinx.ext.napoleon",
    "sphinx.ext.autosectionlabel",
    "sphinx_autodoc_typehints",  # Automatically document param types (less noise in class signature)
    "sphinx_design", # Add e.g. grid layout
    'sphinx_search.extension', # search across multiple docsets in domain
    "nbsphinx",
]

# Use Google docstrings
napoleon_google_docstring = True

# Configure autosummary
autosummary_generate = True  # Turn on sphinx.ext.autosummary
autosummary_ignore_module_all = False # Respect __all__
autodoc_member_order = 'bysource' # Keep original ordering
add_module_names = False  # NB, does not work
autodoc_inherit_docstrings = False # Stops sublcasses from including docs from parent classes

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# Syntax highlighting style
pygments_style = "sphinx"
modindex_common_prefix = ["starsim."]

# List of patterns, relative to source directory, to exclude
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "**.ipynb_checkpoints"]

# Suppress certain warnings
suppress_warnings = ['autosectionlabel.*']


# -- Options for HTML output -------------------------------------------------

html_theme = "pydata_sphinx_theme"
html_theme_options = {
    "collapse_navigation": False,
    "navigation_depth": 3,
    "show_prev_next": True,
    "icon_links": [
        {"name": "Web", "url": "https://starsim.org", "icon": "fas fa-home"},
        {
            "name": "GitHub",
            "url": "https://github.com/starsimhub/starsim",
            "icon": "fab fa-github-square",
        },
    ],
    "navbar_end": ["theme-switcher", "navbar-icon-links"],
    "secondary_sidebar_items": ["navbar-side"],
    "header_links_before_dropdown": 5,
    "footer_start": ["copyright", "footer_start"],
    "footer_end": ["theme-version", "footer_end"],
}
html_sidebars = {
    "**": ["sidebar-nav-bs", "page-toc"],
}
html_logo = "images/idm-logo-transparent.png"
html_favicon = "images/favicon.ico"
html_static_path = ['_static']
html_baseurl = "https://docs.idmod.org/projects/starsim/en/latest"
html_context = {
    'rtd_url': 'https://docs.idmod.org/projects/starsim/en/latest',
    "versions_dropdown": {
        "latest": "devel (latest)",
        "stable": "current (stable)",
    },
    "default_mode": "light",
}
# Add any extra paths that contain custom files
if not on_rtd:
    html_extra_path = ['robots.txt']


# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
html_last_updated_fmt = '%Y-%b-%d'
html_show_sourcelink = True
html_show_sphinx = False
html_copy_source = False
htmlhelp_basename = 'Starsim'

# Add customizations
def setup(app):
    app.add_css_file("theme_overrides.css")


# Modify this to not rerun the Jupyter notebook cells -- usually set by build_docs
nb_ex_default = ['auto', 'never'][0]
nb_ex = os.getenv('NBSPHINX_EXECUTE')
if not nb_ex: nb_ex = nb_ex_default
print(f'\n\nBuilding Jupyter notebooks with build option: {nb_ex}\n\n')
nbsphinx_execute = nb_ex

# OpenSearch options
html_use_opensearch = 'docs.idmod.org/projects/starsim/en/latest'

# -- RTD Sphinx search for searching across the entire domain, default child -------------
if os.environ.get('READTHEDOCS') == 'True':

    search_project_parent = "institute-for-disease-modeling-idm"
    search_project = os.environ["READTHEDOCS_PROJECT"]
    search_version = os.environ["READTHEDOCS_VERSION"]

    rtd_sphinx_search_default_filter = f"subprojects:{search_project}/{search_version}"

    rtd_sphinx_search_filters = {
        "Search this project": f"project:{search_project}/{search_version}",
        "Search all IDM docs": f"subprojects:{search_project_parent}/{search_version}",
    }



import numpy as np
import pylab as pl
import sciris as sc

n = 100
reps = 5
nc = 4
thresh = 1-(0.5/n)
sc.options(dpi=200)
pl.figure(figsize=(12,12))

def matsort(arr):
    s0 = arr.sum(axis=0)
    s1 = arr.sum(axis=1)
    o0 = np.argsort(s0)
    o1 = np.argsort(s1)
    arr = arr[o1,:][:,o0]
    return arr

for rep in range(reps):

    # Truly random
    A = np.random.rand(n,n)
    A = matsort(A)
    
    # Modulo
    b1 = np.random.rand(n,1)
    b2 = np.random.rand(1,n)
    B1 = np.tile(b1, (1,n))
    B2 = np.tile(b2, (n,1))
    B = (B1 + B2) % 1.0
    B = matsort(B)
    
    # Thresholded
    At = A>thresh
    Bt = B>thresh
    Asum = A.sum()/n**2
    Bsum = B.sum()/n**2
    Atsum = (At.sum(axis=1) > 0).sum()
    Btsum = (Bt.sum(axis=1) > 0).sum()
    
    
    #%% Plotting
    pl.subplot(reps,nc,1+nc*rep)
    pl.imshow(A)
    pl.title(f'Truly random {Asum:n}')
    
    pl.subplot(reps,nc,2+nc*rep)
    pl.imshow(B)
    pl.title(f'Modulo {Bsum:n}')
    
    
    pl.subplot(reps,nc,3+nc*rep)
    pl.imshow(At)
    pl.title(f'Truly random {At.sum()}→{Atsum}')
    
    pl.subplot(reps,nc,4+nc*rep)
    pl.imshow(B>thresh)
    pl.title(f'Modulo {Bt.sum()}→{Btsum}')

sc.figlayout()
pl.show()



"""
Vignette 02: "Hello world" multiple STI
DECISIONS:
    - all examples follow similar syntax for specifying more than one disease
TO RESOLVE:
    - what are the default connectors?
"""

import stisim as ss

##################
# @RomeshA
##################
# RA.2.1 Simplest
sim = ss.Sim(modules=[ss.HPV(), ss.Gonorrhea()])

# RA.2.2.1 Changing module pars via the module
hpv_pars = {'eff_condoms':0.5}
hpv = ss.HPV(**hpv_pars)
gonorrhea_pars = {'eff_condoms':0.5}
gonorrhea = ss.Gonorrhea(**gonorrhea_pars)
sim = ss.Sim(modules=[hpv, gonorrhea])

# RA.2.2.2 Changing module pars via the sim par dict
# See RA.1.2.2 in v01_single.py


##################
# @cliffckerr
##################
# CK.2.1 Simplest
sim = ss.Sim(modules=['hiv', 'gonorrhea'])

# CK.2.2.1 Changing module pars via the module
sti_pars = dict(
    n_agents = 10e3,
    start = 1970,
    end = 2020,
)

hiv_pars = dict(
    cd4_pars = dict(
        initial = 500,
        rate = 50,
    ),
)

# Create modules first
hiv = ss.HIV(hiv_pars)
gon = ss.Gonorrhea(transm2f=4.0)

# Create and run the sim
sim = ss.Sim(sti_pars, modules=[hiv, gon])
sim.initialize() # Happens automatically as part of sim.run(), but can call explicitly
assert id(sim.pars.hiv) == id(sim.hiv.pars) # These are the same object after initialization

# CK.2.2.2 Changing module pars via the sim par dict
pars = dict(
    n_agents = 10e3,
    start = 1970,
    end = 2020,
    hiv = dict(
        cd4_pars = dict(
            initial = 500,
            rate = 50,
        ),
    ),
    gonorrhea = dict(
        transm2f = 4.0,
    ),
)

sim = ss.Sim(pars, modules=['hiv', 'gonorrhea'])

##################
# @robynstuart
##################
# RS.2.1 Simplest
sim = ss.Sim(modules=[ss.syphylis(), ss.hiv()])

# RS.2.2.1 Changing module pars via the module
syph_pars = dict(init_prev=0.02)
hiv_pars = dict(init_prev=0.02)
syph = ss.syphilis(pars=syph_pars)
hiv = ss.hiv(pars=hiv_pars)
sim = ss.Sim(modules=[syph, hiv])
sim.initialize()  # Adds modules and parameters
# Parameters can be read like a conditional probability, i.e. prob(transmitting hiv | syphilis infection) = 2
sim['hiv']['syphilis']['rel_trans'] = 4


# RS.2.2.2 Changing module pars via the sim par dict
# See RS.1.2.2


"""
Version and license information.
"""

__all__ = ['__version__', '__versiondate__', '__license__']

__version__ = '2.1.0'
__versiondate__ = '2024-11-07'
__license__ = f'Starsim {__version__} ({__versiondate__}) — © 2023-2024 by IDM'



"""
Test the Dist object from distributions.py
"""

import numpy as np
import sciris as sc
import scipy.stats as sps
import starsim as ss
import matplotlib.pyplot as plt
import pytest

n = 1_000_000 # For testing statistical properties and performance of distributions
m = 5 # For just testing that drawing works
dpy = 365.25 # Number of days per year
sc.options(interactive=False)


def plot_rvs(rvs, times=None, nrows=None):
    fig = plt.figure(figsize=(12,12))
    nrows, ncols = sc.getrowscols(len(rvs), nrows=nrows)
    for i,name,r in rvs.enumitems():
        plt.subplot(nrows, ncols, i+1)
        plt.hist(r.astype(float))
        title = times[name] if times else name
        plt.title(title)
        sc.commaticks()
    sc.figlayout()
    return fig


def make_sim():
    """ Make a tiny sim for initializing the distributions """
    sim = ss.Sim(n_agents=100).init() # Need an empty sim to initialize properly
    return sim


# %% Define the tests
def test_dist(m=m):
    """ Test the Dist class """
    sc.heading('Testing the basic Dist call')
    dist = ss.Dist(distname='random', name='test', strict=False)
    dist.init()
    rvs = dist(m)
    print(rvs)
    assert 0 < rvs.min() < 1, 'Values should be between 0 and 1'

    # Test other options
    dist.show_state()
    dist.plot_hist()
    return rvs


def test_custom_dists(n=n, do_plot=False):
    """ Test all custom dists """
    sc.heading('Testing all custom distributions')

    o     = sc.objdict()
    dists = sc.objdict()
    rvs   = sc.objdict()
    times = sc.objdict()
    for name in ss.dist_list:
        dist_class = getattr(ss, name)
        dist = dist_class(name='test', strict=False)
        dist.init()
        dists[name] = dist
        sc.tic()
        rvs[name] = dist.rvs(n)
        times[name] = sc.toc(name, unit='ms', output='message')
        print(f'{name:10s}: mean = {rvs[name].mean():n}')

    if do_plot:
        plot_rvs(rvs, times=times, nrows=5)

    o.dists = dists
    o.rvs = rvs
    o.times = times
    return o


def test_dists(n=n, do_plot=False):
    """ Test the Dists container """
    sc.heading('Testing Dists container')
    testvals = np.zeros((2,2))

    # Create the objects twice
    for i in range(2):

        # Create a complex object containing various distributions
        obj = sc.prettyobj()
        obj.a = sc.objdict()
        obj.a.mylist = [ss.random(), ss.Dist(distname='uniform', low=2, high=3)]
        obj.b = dict(d3=ss.weibull(c=2), d4=ss.constant(v=0.3))
        dists = ss.Dists(obj).init(sim=make_sim())

        # Call each distribution twice
        for j in range(2):
            rvs = sc.objdict()
            for key,dist in dists.dists.items():
                rvs[str(dist)] = dist(n)
                dist.jump() # Reset
            testvals[i,j] = rvs[0][283] # Pick one of the random values and store it

    # Check that results are as expected
    print(testvals)
    assert np.all(testvals[0,:] == testvals[1,:]), 'Newly initialized objects should match'
    assert np.all(testvals[:,0] != testvals[:,1]), 'After jumping, values should be different'

    if do_plot:
        plot_rvs(rvs)

    o = sc.objdict()
    o.dists = dists
    o.rvs = rvs
    return dists


def test_scipy(m=m):
    """ Test that SciPy distributions also work """
    sc.heading('Testing SciPy distributions')

    # Make SciPy distributions in two different ways
    dist1 = ss.Dist(dist=sps.expon, name='scipy', scale=2, strict=False).init() # Version 1: callable
    dist2 = ss.Dist(dist=sps.expon(scale=2), name='scipy', strict=False).init() # Version 2: frozen
    rvs1 = dist1(m)
    rvs2 = dist2(m)

    # Check that they match
    print(rvs1)
    assert np.array_equal(rvs1, rvs2), 'Arrays should match'

    return dist1, dist2


def test_exceptions(m=m):
    """ Check that exceptions are being appropriately raised """
    sc.heading('Testing exceptions and strict')

    # Create a strict distribution
    dist = ss.random(strict=True, auto=False)
    with pytest.raises(ss.distributions.DistNotInitializedError):
        dist(m) # Check that we can't call an uninitialized

    # Initialize and check we can't call repeatedly
    dist.init(trace='test', sim=make_sim())
    rvs = dist(m)
    with pytest.raises(ss.distributions.DistNotReadyError):
        dist(m) # Check that we can't call an already-used distribution

    # Check that we can with a non-strict Dist
    dist2 = ss.random(strict=False)
    dist2.init(trace='test')
    rvs2 = sc.autolist()
    for i in range(2):
        rvs2 += dist2(m) # We should be able to call multiple times with no problem

    print(rvs)
    print(rvs2)
    assert np.array_equal(rvs, rvs2[0]), 'Separate dists should match'
    assert not np.array_equal(rvs2[0], rvs2[1]), 'Multiple calls to the same dist should not match'

    return dist, dist2


def test_reset(m=m):
    """ Check that reset works as expected """
    sc.heading('Testing reset')

    # Create and draw two sets of random numbers
    dist = ss.random(seed=533, strict=False).init()
    r1 = dist.rvs(m)
    r2 = dist.rvs(m)
    assert all(r1 != r2)

    # Reset to the most recent state
    dist.reset(-1)
    r3 = dist.rvs(m)
    assert all(r3 == r2)

    # Reset to the initial state
    dist.reset(0)
    r4 = dist.rvs(m)
    assert all(r4 == r1)

    for r in [r1, r2, r3, r4]:
        print(r)

    return dist


def make_fake_sim(n=10):
    """ Define a fake sim object for testing slots """
    sim = sc.prettyobj()
    sim.n = n
    sim.people = sc.prettyobj()
    sim.people.uid = np.arange(sim.n)
    sim.people.slot = np.arange(sim.n)
    sim.people.age = np.random.uniform(0, 100, size=sim.n)
    return sim


def test_callable(n=n):
    """ Test callable parameters """
    sc.heading('Testing a uniform distribution with callable parameters')

    sim = make_fake_sim()

    # Define a parameter as a function
    def custom_loc(module, sim, uids):
        out = sim.people.age[uids]
        return out

    scale = 1
    d1 = ss.normal(name='callable', loc=custom_loc, scale=scale).init(sim=sim)
    d2 = ss.lognorm_ex(name='callable', mean=custom_loc, std=scale).init(sim=sim)

    uids = np.array([1, 3, 7, 9])
    draws1 = d1.rvs(uids)
    draws2 = d2.rvs(uids)
    print(f'Input ages were: {sim.people.age[uids]}')
    print(f'Output samples were: {draws1}, {draws2}')

    for draws in [draws1, draws2]:
        meandiff = np.abs(sim.people.age[uids] - draws).mean()
        assert meandiff < scale*3, 'Outputs should match ages'

    return d1


def test_array(n=n):
    """ Test array parameters """
    sc.heading('Testing uniform with a array parameters')

    uids = np.array([1, 3])
    low  = np.array([1, 100]) # Low
    high = np.array([3, 125]) # High

    d = ss.uniform(low=low, high=high, strict=False).init(slots=np.arange(uids.max()+1))
    draws = d.rvs(uids)
    print(f'Uniform sample for uids {uids} returned {draws}')

    assert len(draws) == len(uids)
    for i in range(len(uids)):
        assert low[i] < draws[i] < low[i] + high[i], 'Invalid value'
    return draws


def test_repeat_slot():
    """ Test behavior of repeated slots """
    sc.heading('Test behavior of repeated slots')

    # Initialize parameters
    slots = np.array([4,2,3,2,2,3])
    n = len(slots)
    uids = np.arange(n)
    low = np.arange(n)
    high = low + 1

    # Draw values
    d = ss.uniform(low=low, high=high, strict=False).init(slots=slots)
    draws = d.rvs(uids)

    # Print and test
    print(f'Uniform sample for slots {slots} returned {draws}')
    assert len(draws) == len(slots)

    unique_slots = np.unique(slots)
    for s in unique_slots:
        inds = np.where(slots==s)[0]
        frac, integ = np.modf(draws[inds])
        assert np.allclose(integ, low[inds]), 'Integral part should match the low parameter'
        assert np.allclose(frac, frac[0]), 'Same random numbers, so should be same fractional part'
    return draws


def test_timepar_dists():
    """ Test interaction of distributions and timepars """
    sc.heading('Test interaction of distributions and timepars')

    # Set parameters
    n = int(10e3)
    u1 = 'day'
    u2 = 'year'
    dt_dur = 0.1
    dt_rate = 20.0
    ratio_dur  = ss.time_ratio(u1, 1.0, u2, dt_dur)
    ratio_rate = ss.time_ratio(u2, 1.0, u1, dt_rate)

    # Create time parameters
    v = sc.objdict()
    v.base = 30.0
    v.dur = ss.dur(v.base, unit=u1, parent_unit=u2, parent_dt=dt_dur).init()
    v.rate = ss.rate(v.base, unit=u2, parent_unit=u1, parent_dt=dt_rate).init() # Swap units

    # Check distributions that scale linearly with time, with the parameter we'll set
    print('Testing linear distributions ...')
    linear_dists = dict(
        constant   = 'v',
        uniform    = 'high',
        normal     = 'loc',
        lognorm_ex = 'mean',
        expon      = 'scale',
        weibull    = 'scale',
        gamma      = 'scale',
    )

    for name,par in linear_dists.items():
        dist_class = getattr(ss, name)

        # Create the dists, the first parameter of which should have time units
        dists = sc.objdict()
        for key,val in v.items():
            pardict = {par:val} # Convert to a tiny dictionary to insert the correct name
            dists[key] = dist_class(**pardict, name=key, strict=False).init()

        # Create th erandom variates
        rvs = sc.objdict()
        for k,dist in dists.items():
            rvs[k] = dist.rvs(n)

        # Check that the distributions match
        rtol = 0.1 # Be somewhat generous with the uncertainty
        expected = rvs.base.mean()
        expected_dur = expected*ratio_dur
        expected_rate = expected/ratio_rate
        actual_dur = rvs.dur.mean()
        actual_rate = rvs.rate.mean()
        assert np.isclose(expected_dur, actual_dur, rtol=rtol), f'Duration not close for {name}: {expected_dur:n} ≠ {actual_dur:n}'
        assert np.isclose(expected_rate, actual_rate, rtol=rtol), f'Rate not close for {name}: {expected_rate:n} ≠ {actual_rate:n}'
        sc.printgreen(f'✓ {name} passed: {expected_dur:n} ≈ {actual_dur:n}')

    # Check that unitless distributions fail
    print('Testing unitless distributions ...')
    par = ss.dur(10)
    unitless_dists = ['lognorm_im', 'randint', 'choice']
    for name in unitless_dists:
        dist_class = getattr(ss, name)
        with pytest.raises(NotImplementedError):
            dist = dist_class(par, name='notimplemented', strict=False).init()
            dist.rvs(n)
        sc.printgreen(f'✓ {name} passed: raised appropriate error')

    # Check special distributions
    print('Testing Poisson distribution ...')
    lam1 = ss.rate(dpy, 'year', parent_unit='year').init()
    lam2 = ss.rate(1,   'day',  parent_unit='year').init()
    poi1 = ss.poisson(lam=lam1, strict=False).init()
    poi2 = ss.poisson(lam=lam2, strict=False).init()
    mean1 = poi1.rvs(n).mean()
    mean2 = poi2.rvs(n).mean()
    assert np.isclose(mean1, mean2, rtol=rtol), f'Poisson values do not match for {lam1} and {lam2}: {mean1:n} ≠ {mean2:n}'
    sc.printgreen(f'✓ poisson passed: {mean1:n} ≈ {mean2:n}')

    print('Testing Bernoulli distribution ...')
    p1 = 0.1
    p2 = ss.time_prob(0.01, parent_dt=10).init()
    ber1 = ss.bernoulli(p=p1, strict=False).init()
    ber2 = ss.bernoulli(p=p2, strict=False).init()
    mean1 = ber1.rvs(n).mean()
    mean2 = ber2.rvs(n).mean()
    assert np.isclose(mean1, mean2, rtol=rtol), f'Bernoulli values do not match for {lam1} and {lam2}: {mean1:n} ≠ {mean2:n}'
    sc.printgreen(f'✓ bernoulli passed: {mean1:n} ≈ {mean2:n}')

    return ber2


def test_timepar_callable():
    sc.heading('Test that timepars work with (some) callable functions')

    print('Testing callable parameters with regular-scaling distributions')
    sim = make_fake_sim(n=10)
    uids = np.array([1, 3, 7, 9])

    def age_year(module, sim, uids):
        """ Extract age as a year """
        out = sim.people.age[uids].copy()
        return out

    def age_day(module, sim, uids):
        """ Extract age as a day """
        out = sim.people.age[uids].copy()
        out *= dpy # Convert to days manually
        return out

    scale = 1e-3 # Set to a very small but nonzero scale
    loc  = ss.dur(age_year, unit='year', parent_unit='day').init(update_values=False)
    mean = ss.dur(age_day,  unit='day',  parent_unit='day').init(update_values=False)
    d3 = ss.normal(name='callable', loc=loc, scale=scale).init(sim=sim)
    d4 = ss.lognorm_ex(name='callable', mean=mean, std=scale).init(sim=sim)
    draws3 = d3.rvs(uids)
    draws4 = d4.rvs(uids)
    age_in_days = sim.people.age[uids]*dpy
    draw_diff = np.abs(draws3 - draws4).mean()
    age_diff = np.abs(age_in_days - draws3).mean()
    print(f'Input ages were:\n{sim.people.age[uids]}')
    print(f'Output samples were:\n{sc.sigfiground(draws3)}\n{sc.sigfiground(draws4)}')
    assert draw_diff < 1, 'Day and year outputs should match to the nearest day'
    assert age_diff < 1, 'Distribution outputs should match ages to the nearest day'

    print('Testing callable parameters with Bernoulli distributions')
    n = 100_000
    sim = make_fake_sim(n=n)
    uids = np.random.choice(n, size=n//2, replace=False)
    age = sim.people.age[uids]
    mean = age.mean()
    young = sc.findinds(age<=mean)
    old = sc.findinds(age>mean)
    p_young = 0.001
    p_old = 0.010

    def age_prob(module, sim, uids):
        out = np.zeros_like(age)
        out[young] = p_young
        out[old]   = p_old
        return out

    parent_dt = 10
    p1 = age_prob
    p2 = ss.time_prob(age_prob, parent_dt=parent_dt).init(update_values=False)
    ber1 = ss.bernoulli(name='base', p=p1).init(sim=sim)
    ber2 = ss.bernoulli(name='time', p=p2).init(sim=sim)
    brvs1 = ber1.rvs(uids)
    brvs2 = ber2.rvs(uids)

    rtol = 0.5 # We're dealing with order-of-magnitude differences but small numbers, so be generous to avoid random failures
    sum1 = brvs1.sum()*parent_dt
    sum2 = brvs2.sum()
    assert np.isclose(sum1, sum2, rtol=rtol), f'Callable Bernoulli sums did not match: {sum1}  ≠  {sum2}'
    sc.printgreen(f'✓ Callable Bernoulli sums matched: {sum1:n} ≈ {sum2:n}')
    for key,expected,inds in zip(['young', 'old'], [p_young, p_old], [young, old]):
        m1 = brvs1[inds].mean()
        m2 = brvs2[inds].mean()/parent_dt
        assert np.allclose([expected, m1], [expected, m2], rtol=rtol), f'Callable Bernoulli proportions did not match: {expected:n}  ≠  {m1:n}  ≠  {m2:n}'
        sc.printgreen(f'✓ Callable Bernoulli proportions matched: {expected:n}  ≈  {m1:n}  ≈  {m2:n}')

    return


# %% Run as a script
if __name__ == '__main__':
    do_plot = True
    sc.options(interactive=do_plot)

    T = sc.timer()

    o1 = test_dist()
    o2 = test_custom_dists(do_plot=do_plot)
    o3 = test_dists(do_plot=do_plot)
    o4 = test_scipy()
    o5 = test_exceptions()
    o6 = test_reset()
    o7 = test_callable()
    o8 = test_array()
    o9 = test_repeat_slot()
    o10 = test_timepar_dists()
    o10 = test_timepar_callable()

    T.toc()



"""
Vignette 07: Custom connectors
TO RESOLVE:
- do we need a custom connector class?
    - any other example for how to do this?
"""
import stisim as ss


##################
# @pausz
##################

# Create instances of STIsim modules
hiv = ss.HIV()
gonorrhea = ss.Gonorrhea()
modules = [hiv, gonorrhea]

# Create a connector with the two modules
connector = ss.Connector([hiv, gonorrhea])

# Define interactions between the modules
def custom_interaction_1():
    pass
def custom_interaction_2():
    pass

connector.add_interaction('interaction_1', source_module, target_module, custom_interaction_1)
connector.add_interaction('interaction_2', source_module, target_module, custom_interaction_2)
# The connector class would have a method apply_interactions() called by sim.

# Run the simulation using the connected modules
sim = ss.Sim(modules=[hiv, gonorrhea], connectors=connector)

##################
# @robynstuart
##################

# Case 1: modifying simple connectors
syph_pars = dict(init_prev=0.02)
hiv_pars = dict(
    init_prev=0.02,
    # Here come the connectors - simple connectors are multipliers on any people attribute (e.g. people.hiv.rel_trans)
    #   - syphilis increases risk of acquiring HIV x2
    #   - syphilis increases risk of transmitting HIV x3
    #   - syphilis increases HIV viral load x10
    syphilis={'rel_trans': 2, 'rel_sus': 3, 'viral_load': 10},
)
syph = ss.syphilis(pars=syph_pars)
hiv = ss.syphilis(pars=hiv_pars)
sim = ss.Sim(modules=[syph, hiv])
sim.initialize()  # Adds modules and parameters
# Parameters can be read like a conditional probability, i.e. prob(transmitting hiv | syphilis infection) = 2
sim['hiv']['syphilis']['rel_trans'] = 4

# Case 2: creating complex connectors
# Example where rel_trans is higher for those with lower cd4 count...
complex_connector = lambda rel_trans, cd4: (6*rel_trans**2 * (cd4<500)) + (2*rel_trans * (cd4>500))
hiv_pars = dict(
    init_prev=0.02,
    syphilis={'rel_trans': complex_connector},
)
syph = ss.syphilis(pars=syph_pars)
hiv = ss.hiv(pars=hiv_pars)
sim = ss.Sim(modules=[syph, hiv])

# Case 3: creating even more complex connectors
# Completely made-up example where HIV makes syphilis incurable and permanently latent.
# At this stage, I feel that most examples of connectors could be managed by modifying the people
# attributes, so this would be a relatively rare example. If it became more common we could include
# defaults. I think that the intervention class would be able to handle this as-is, but for clarity
# we could make a Connector subclass.
class hiv_syph_effect(ss.Intervention):
    def __init__(self):
        pass
    def initialize(self, sim):
        pass
    def apply(self, sim):
        hiv_syph_inds = sim.people.hiv.infected * sim.people.syphilis.infected
        sim.people.syphilis.dur_latent[hiv_syph_inds] = 100
        sim.people.syphilis.treatable[hiv_syph_inds] = False


# Case 4: creating even MORE complex connectors
# For anything more complex than this, modify the module directly, i.e.
class complex_hiv_syphilis(ss.HIV):
    # Write your own!
    pass




"""
Define the calibration class
"""
import os
import numpy as np
import pandas as pd
import sciris as sc
import optuna as op
import matplotlib.pyplot as plt
import starsim as ss


__all__ = ['Calibration', 'compute_gof']


def compute_gof(actual, predicted, normalize=True, use_frac=False, use_squared=False,
                as_scalar='none', eps=1e-9, skestimator=None, estimator=None, **kwargs):
    """
    Calculate the goodness of fit. By default use normalized absolute error, but
    highly customizable. For example, mean squared error is equivalent to
    setting normalize=False, use_squared=True, as_scalar='mean'.

    Args:
        actual      (arr):   array of actual (data) points
        predicted   (arr):   corresponding array of predicted (model) points
        normalize   (bool):  whether to divide the values by the largest value in either series
        use_frac    (bool):  convert to fractional mismatches rather than absolute
        use_squared (bool):  square the mismatches
        as_scalar   (str):   return as a scalar instead of a time series: choices are sum, mean, median
        eps         (float): to avoid divide-by-zero
        skestimator (str):   if provided, use this scikit-learn estimator instead
        estimator   (func):  if provided, use this custom estimator instead
        kwargs      (dict):  passed to the scikit-learn or custom estimator

    Returns:
        gofs (arr): array of goodness-of-fit values, or a single value if as_scalar is True

    **Examples**::

        x1 = np.cumsum(np.random.random(100))
        x2 = np.cumsum(np.random.random(100))

        e1 = compute_gof(x1, x2) # Default, normalized absolute error
        e2 = compute_gof(x1, x2, normalize=False, use_frac=False) # Fractional error
        e3 = compute_gof(x1, x2, normalize=False, use_squared=True, as_scalar='mean') # Mean squared error
        e4 = compute_gof(x1, x2, skestimator='mean_squared_error') # Scikit-learn's MSE method
        e5 = compute_gof(x1, x2, as_scalar='median') # Normalized median absolute error -- highly robust
    """

    # Handle inputs
    actual    = np.array(sc.dcp(actual), dtype=float)
    predicted = np.array(sc.dcp(predicted), dtype=float)

    # Scikit-learn estimator is supplied: use that
    if skestimator is not None: # pragma: no cover
        try:
            import sklearn.metrics as sm
            sklearn_gof = getattr(sm, skestimator) # Shortcut to e.g. sklearn.metrics.max_error
        except ImportError as E:
            errormsg = f'You must have scikit-learn >=0.22.2 installed: {str(E)}'
            raise ImportError(errormsg) from E
        except AttributeError as E:
            errormsg = f'Estimator {skestimator} is not available; see https://scikit-learn.org/stable/modules/model_evaluation.html#scoring-parameter for options'
            raise AttributeError(errormsg) from E
        gof = sklearn_gof(actual, predicted, **kwargs)
        return gof

    # Custom estimator is supplied: use that
    if estimator is not None: # pragma: no cover
        try:
            gof = estimator(actual, predicted, **kwargs)
        except Exception as E:
            errormsg = f'Custom estimator "{estimator}" must be a callable function that accepts actual and predicted arrays, plus optional kwargs'
            raise RuntimeError(errormsg) from E
        return gof

    # Default case: calculate it manually
    else:
        # Key step -- calculate the mismatch!
        gofs = abs(np.array(actual) - np.array(predicted))

        if normalize and not use_frac:
            actual_max = abs(actual).max()
            if actual_max > 0:
                gofs /= actual_max

        if use_frac:
            if (actual<0).any() or (predicted<0).any():
                print('Warning: Calculating fractional errors for non-positive quantities is ill-advised!')
            else:
                maxvals = np.maximum(actual, predicted) + eps
                gofs /= maxvals

        if use_squared:
            gofs = gofs**2

        if as_scalar == 'sum':
            gofs = np.sum(gofs)
        elif as_scalar == 'mean':
            gofs = np.mean(gofs)
        elif as_scalar == 'median':
            gofs = np.median(gofs)

        return gofs


class Calibration(sc.prettyobj): # pragma: no cover
    """
    A class to handle calibration of Starsim simulations. Uses the Optuna hyperparameter
    optimization library (optuna.org).

    Args:
        sim          (Sim)  : the simulation to calibrate
        data         (df)   : pandas dataframe (or dataframe-compatible dict) of the data to calibrate to
        calib_pars   (dict) : a dictionary of the parameters to calibrate of the format dict(key1=[best, low, high])
        n_trials     (int)  : the number of trials per worker
        n_workers    (int)  : the number of parallel workers (default: maximum number of available CPUs)
        total_trials (int)  : if n_trials is not supplied, calculate by dividing this number by n_workers
        reseed       (bool) : whether to generate new random seeds for each trial
        weights      (dict) : the relative weights of each data source
        fit_args     (dict) : a dictionary of options that are passed to sim.compute_fit() to calculate the goodness-of-fit
        sep          (str)  : the separate between different types of results, e.g. 'hiv.deaths' vs 'hiv_deaths'
        name         (str)  : the name of the database (default: 'starsim_calibration')
        db_name      (str)  : the name of the database file (default: 'starsim_calibration.db')
        keep_db      (bool) : whether to keep the database after calibration (default: false)
        storage      (str)  : the location of the database (default: sqlite)
        rand_seed    (int)  : if provided, use this random seed to initialize Optuna runs (for reproducibility)
        label        (str)  : a label for this calibration object
        die          (bool) : whether to stop if an exception is encountered (default: false)
        debug        (bool) : if True, do not run in parallel
        verbose      (bool) : whether to print details of the calibration

    Returns:
        A Calibration object
    """
    def __init__(self, sim, data, calib_pars, n_trials=None, n_workers=None, total_trials=None, reseed=True,
                 weights=None, fit_args=None, sep='.', name=None, db_name=None, keep_db=None, storage=None,
                 rand_seed=None, sampler=None, label=None, die=False, debug=False, verbose=True, save_results=False):

        # Handle run arguments
        if n_trials  is None: n_trials  = 20
        if n_workers is None: n_workers = sc.cpu_count()
        if name      is None: name      = 'starsim_calibration'
        if db_name   is None: db_name   = f'{name}.db'
        if keep_db   is None: keep_db   = False
        if storage   is None: storage   = f'sqlite:///{db_name}'
        if total_trials is not None: n_trials = int(np.ceil(total_trials/n_workers))
        kw = dict(n_trials=int(n_trials), n_workers=int(n_workers), debug=debug, name=name, db_name=db_name,
                  keep_db=keep_db, storage=storage, rand_seed=rand_seed, sampler=sampler)
        self.run_args = sc.objdict(kw)

        # Handle other inputs
        self.label      = label
        self.sim        = sim
        self.calib_pars = calib_pars
        self.reseed     = reseed
        self.sep        = sep
        self.weights    = sc.mergedicts(weights)
        self.fit_args   = sc.mergedicts(fit_args)
        self.die        = die
        self.verbose    = verbose
        self.save_results = save_results
        self.calibrated = False
        self.before_sim = None
        self.after_sim  = None

        # Load data -- this is expecting a dataframe with a column for 'time' and other columns for to sim results
        self.data = ss.validate_sim_data(data, die=True)

        # Temporarily store a filename
        self.tmp_filename = 'tmp_calibration_%05i.obj'

        # Initialize sim
        if not self.sim.initialized:
            self.sim.init()

        # Figure out which sim results to get
        self.sim_result_list = self.data.cols

        return

    def run_sim(self, calib_pars=None, label=None):
        """ Create and run a simulation """
        sim = sc.dcp(self.sim)
        if label: sim.label = label

        sim = self.translate_pars(sim, calib_pars=calib_pars)

        # Run the sim
        try:
            sim.run()
            return sim

        except Exception as E:
            if self.die:
                raise E
            else:
                print(f'Encountered error running sim!\nParameters:\n{calib_pars}\nTraceback:\n{sc.traceback()}')
                output = None
                return output

    @staticmethod
    def translate_pars(sim=None, calib_pars=None):
        """ Take the nested dict of calibration pars and modify the sim """

        if 'rand_seed' in calib_pars:
            sim.pars['rand_seed'] = calib_pars['rand_seed']

        for parname, spec in calib_pars.items():
            if parname == 'rand_seed':
                continue

            if 'path' not in spec:
                raise ValueError(f'Cannot map {parname} because "path" is missing from the parameter configuration.')

            p = spec['path']

            if len(p) != 3:
                raise ValueError(f'Cannot map {parname} because "path" must be a tuple of length 3.')

            modtype = p[0]
            dkey = p[1]
            dparkey = p[2]
            dparval = spec['value']
            targetpar = sim[modtype][dkey].pars[dparkey]

            if sc.isnumber(targetpar):
                sim[modtype][dkey].pars[dparkey] = dparval
            elif isinstance(targetpar, ss.Dist):
                sim[modtype][dkey].pars[dparkey].set(dparval)
            else:
                errormsg = 'Type not implemented'
                raise ValueError(errormsg)

        return sim

    def trial_to_sim_pars(self, pardict=None, trial=None):
        """
        Take in an optuna trial and sample from pars, after extracting them from the structure they're provided in

        Different use cases:
            - pardict is self.calib_pars, i.e. {'diseases':{'hiv':{'art_efficacy':[0.96, 0.9, 0.99]}}}, need to sample
            - pardict is self.initial_pars, i.e. {'diseases':{'hiv':{'art_efficacy':[0.96, 0.9, 0.99]}}}, pull 1st vals
            - pardict is self.best_pars, i.e. {'diseases':{'hiv':{'art_efficacy':0.96786}}}, pull single vals
        """
        pars = sc.dcp(pardict)
        for parname, spec in pars.items():

            if 'value' in spec:
                # Already have a value, likely running initial or final values as part of checking the fit
                continue

            if 'sampler' in spec:
                sampler = spec.pop('sampler')
                sampler_fn = getattr(trial, sampler)
            else:
                sampler_fn = trial.suggest_float

            path = spec.pop('path', None) # remove path
            guess = spec.pop('guess', None) # remove guess
            spec['value'] = sampler_fn(name=parname, **spec) # Sample!
            spec['path'] = path
            spec['guess'] = guess

        return pars

    @staticmethod
    def sim_to_df(sim): # TODO: remove this method
        """ Convert a sim to the expected dataframe type """
        df_res = sim.to_df(sep='.')
        df_res['t'] = df_res['timevec']
        df_res = df_res.set_index('t')
        df_res['time'] = np.floor(np.round(df_res.index, 1)).astype(int)
        return df_res

    def run_trial(self, trial):
        """ Define the objective for Optuna """
        if self.calib_pars is not None:
            calib_pars = self.trial_to_sim_pars(self.calib_pars, trial)
        else:
            calib_pars = None

        if self.reseed:
            calib_pars['rand_seed'] = trial.suggest_int('rand_seed', 0, 1_000_000) # Choose a random rand_seed

        sim = self.run_sim(calib_pars)

        # Export results # TODO: make more robust
        df_res = self.sim_to_df(sim)
        sim_results = sc.objdict()

        for skey in self.sim_result_list:
            if 'prevalence' in skey or skey.startswith('n_'):
                model_output = df_res.groupby(by='time')[skey].mean()
            else:
                model_output = df_res.groupby(by='time')[skey].sum()
            sim_results[skey] = model_output.values

        sim_results['time'] = model_output.index.values
        # Store results in temporary files
        if self.save_results:
            filename = self.tmp_filename % trial.number
            sc.save(filename, sim_results)

        # Compute fit
        fit = self.compute_fit(df_res=df_res)
        return fit

    def compute_fit(self, sim=None, df_res=None):
        """ Compute goodness-of-fit """
        fit = 0

        # TODO: reduce duplication with above
        if df_res is None:
            df_res = self.sim_to_df(sim)
        for skey in self.sim_result_list:
            if 'prevalence' in skey or skey.startswith('n_'):
                model_output = df_res.groupby(by='time')[skey].mean()
            else:
                model_output = df_res.groupby(by='time')[skey].sum()

            data = self.data[skey]
            combined = pd.merge(data, model_output, how='left', on='time')
            combined['diffs'] = combined[skey+'_x'] - combined[skey+'_y']
            gofs = compute_gof(combined.dropna()[skey+'_x'], combined.dropna()[skey+'_y'])

            losses = gofs  #* self.weights[skey]
            mismatch = losses.sum()
            fit += mismatch

        return fit

    def worker(self):
        """ Run a single worker """
        if self.verbose:
            op.logging.set_verbosity(op.logging.DEBUG)
        else:
            op.logging.set_verbosity(op.logging.ERROR)
        study = op.load_study(storage=self.run_args.storage, study_name=self.run_args.name, sampler = self.run_args.sampler)
        output = study.optimize(self.run_trial, n_trials=self.run_args.n_trials, callbacks=None)
        return output

    def run_workers(self):
        """ Run multiple workers in parallel """
        if self.run_args.n_workers > 1 and not self.run_args.debug: # Normal use case: run in parallel
            output = sc.parallelize(self.worker, iterarg=self.run_args.n_workers)
        else: # Special case: just run one
            output = [self.worker()]
        return output

    def remove_db(self):
        """ Remove the database file if keep_db is false and the path exists """
        try:
            if 'sqlite' in self.run_args.storage:
                # Delete the file from disk
                if os.path.exists(self.run_args.db_name):
                    os.remove(self.run_args.db_name)
                if self.verbose: print(f'Removed existing calibration file {self.run_args.db_name}')
            else:
                # Delete the study from the database e.g., mysql
                op.delete_study(study_name=self.run_args.name, storage=self.run_args.storage)
                if self.verbose: print(f'Deleted study {self.run_args.name} in {self.run_args.storage}')
        except Exception as E:
            if self.verbose:
                print('Could not delete study, skipping...')
                print(str(E))
        return

    def make_study(self):
        """ Make a study, deleting one if it already exists """
        if not self.run_args.keep_db:
            self.remove_db()
        if self.run_args.rand_seed is not None:
            sampler = op.samplers.RandomSampler(self.run_args.rand_seed)
            sampler.reseed_rng()
            raise NotImplementedError('Implemented but does not work')
        else:
            sampler = None
        if self.verbose: print(self.run_args.storage)
        output = op.create_study(storage=self.run_args.storage, study_name=self.run_args.name, sampler=sampler)
        return output

    def calibrate(self, calib_pars=None, confirm_fit=False, load=False, tidyup=True, **kwargs):
        """
        Perform calibration.

        Args:
            calib_pars (dict): if supplied, overwrite stored calib_pars
            confirm_fit (bool): if True, run simulations with parameters from before and after calibration
            load (bool): whether to load existing trials from the database (if rerunning the same calibration)
            tidyup (bool): whether to delete temporary files from trial runs
            verbose (bool): whether to print output from each trial
            kwargs (dict): if supplied, overwrite stored run_args (n_trials, n_workers, etc.)
        """
        # Load and validate calibration parameters
        if calib_pars is not None:
            self.calib_pars = calib_pars
        self.run_args.update(kwargs) # Update optuna settings

        # Run the optimization
        t0 = sc.tic()
        self.make_study()
        self.run_workers()
        study = op.load_study(storage=self.run_args.storage, study_name=self.run_args.name, sampler = self.run_args.sampler)
        self.best_pars = sc.objdict(study.best_params)
        self.elapsed = sc.toc(t0, output=True)

        self.sim_results = []
        if load:
            if self.verbose: print('Loading saved results...')
            for trial in study.trials:
                n = trial.number
                try:
                    filename = self.tmp_filename % trial.number
                    results = sc.load(filename)
                    self.sim_results.append(results)
                    if tidyup:
                        try:
                            os.remove(filename)
                            if self.verbose: print(f'    Removed temporary file {filename}')
                        except Exception as E:
                            errormsg = f'Could not remove {filename}: {str(E)}'
                            if self.verbose: print(errormsg)
                    if self.verbose: print(f'  Loaded trial {n}')
                except Exception as E:
                    errormsg = f'Warning, could not load trial {n}: {str(E)}'
                    if self.verbose: print(errormsg)

        # Compare the results
        self.parse_study(study)

        if self.verbose: print('Best pars:', self.best_pars)

        # Tidy up
        self.calibrated = True
        if not self.run_args.keep_db:
            self.remove_db()

        # Optionally compute the sims before and after the fit
        if confirm_fit:
            self.confirm_fit()

        return self

    def confirm_fit(self):
        """ Run before and after simulations to validate the fit """

        if self.verbose: print('\nConfirming fit...')

        before_pars = sc.dcp(self.calib_pars)
        for spec in before_pars.values():
            spec['value'] = spec['guess'] # Use guess values

        after_pars = sc.dcp(self.calib_pars)
        for parname, spec in after_pars.items():
            spec['value'] = self.best_pars[parname]

        self.before_sim = self.run_sim(calib_pars=before_pars, label='Before calibration')
        self.after_sim  = self.run_sim(calib_pars=after_pars, label='After calibration')
        self.before_fit = self.compute_fit(self.before_sim)
        self.after_fit  = self.compute_fit(self.after_sim)

        # Add the data to the sims
        for sim in [self.before_sim, self.after_sim]:
            sim.init_data(self.data)

        print(f'Fit with original pars: {self.before_fit:n}')
        print(f'Fit with best-fit pars: {self.after_fit:n}')
        if self.after_fit <= self.before_fit:
            print('✓ Calibration improved fit')
        else:
            print('✗ Calibration did not improve fit, but this sometimes happens stochastically and is not necessarily an error')

        return self.before_fit, self.after_fit

    def parse_study(self, study):
        """Parse the study into a data frame -- called automatically """
        best = study.best_params
        self.best_pars = best

        if self.verbose: print('Making results structure...')
        results = []
        n_trials = len(study.trials)
        failed_trials = []
        for trial in study.trials:
            data = {'index':trial.number, 'mismatch': trial.value}
            for key,val in trial.params.items():
                data[key] = val
            if data['mismatch'] is None:
                failed_trials.append(data['index'])
            else:
                results.append(data)
        if self.verbose: print(f'Processed {n_trials} trials; {len(failed_trials)} failed')

        keys = ['index', 'mismatch'] + list(best.keys())
        data = sc.objdict().make(keys=keys, vals=[])
        for i,r in enumerate(results):
            for key in keys:
                if key not in r:
                    warnmsg = f'Key {key} is missing from trial {i}, replacing with default'
                    print(warnmsg)
                    r[key] = best[key]
                data[key].append(r[key])
        self.study_data = data
        self.df = sc.dataframe.from_dict(data)
        self.df = self.df.sort_values(by=['mismatch']) # Sort
        return

    def to_json(self, filename=None, indent=2, **kwargs):
        """ Convert the results to JSON """
        order = np.argsort(self.df['mismatch'])
        json = []
        for o in order:
            row = self.df.iloc[o,:].to_dict()
            rowdict = dict(index=row.pop('index'), mismatch=row.pop('mismatch'), pars={})
            for key,val in row.items():
                rowdict['pars'][key] = val
            json.append(rowdict)
        self.json = json
        if filename:
            return sc.savejson(filename, json, indent=indent, **kwargs)
        else:
            return json

    def plot_sims(self, **kwargs):
        """
        Plot sims, before and after calibration.

        Args:
            kwargs (dict): passed to MultiSim.plot()
        """
        if self.before_sim is None:
            self.confirm_fit()
        msim = ss.MultiSim([self.before_sim, self.after_sim])
        fig = msim.plot(**kwargs)
        return fig

    def plot_trend(self, best_thresh=None, fig_kw=None):
        """
        Plot the trend in best mismatch over time.

        Args:
            best_thresh (int): Define the threshold for the "best" fits, relative to the lowest mismatch value (if None, show all)
            fig_kw (dict): passed to plt.figure()
        """
        df = self.df.sort_values('index') # Make a copy of the dataframe, sorted by trial number
        mismatch = sc.dcp(df['mismatch'].values)
        best_mismatch = np.zeros(len(mismatch))
        for i in range(len(mismatch)):
            best_mismatch[i] = mismatch[:i+1].min()
        smoothed_mismatch = sc.smooth(mismatch)
        fig = plt.figure(**sc.mergedicts(fig_kw))

        ax1 = plt.subplot(2,1,1)
        plt.plot(mismatch, alpha=0.2, label='Original')
        plt.plot(smoothed_mismatch, lw=3, label='Smoothed')
        plt.plot(best_mismatch, lw=3, label='Best')

        ax2 = plt.subplot(2,1,2)
        max_mismatch = mismatch.min()*best_thresh if best_thresh is not None else np.inf
        inds = sc.findinds(mismatch<=max_mismatch)
        plt.plot(best_mismatch, lw=3, label='Best')
        plt.scatter(inds, mismatch[inds], c=mismatch[inds], label='Trials')
        for ax in [ax1, ax2]:
            plt.sca(ax)
            plt.grid(True)
            plt.legend()
            sc.setylim()
            sc.setxlim()
            plt.xlabel('Trial number')
            plt.ylabel('Mismatch')
        sc.figlayout()
        return fig


"""
Run simplest tests
"""

# %% Imports and settings
import starsim as ss
import matplotlib.pyplot as plt


ppl = ss.People(9000)
ppl.networks = ss.ndict(ss.MFNet(), ss.MaternalNet())

hiv = ss.HIV()
hiv.pars['beta'] = {'mf': [0.08, 0.04], 'maternal': [0.2, 0]}

ng = ss.Gonorrhea()

sim = ss.Sim(people=ppl, demographics=[ss.Pregnancy()], diseases=[hiv, ng])
sim.initialize()
sim.run()

# Plotting
fig, axv = plt.subplots(3,1, figsize=(10,8))
for ax, module in zip(axv[:2], ['HIV', 'Gonorrhea']):
    results = sim.results[module.lower()]
    for ch in results.keys():
        if ch[:2] == 'n_':
            ax.plot(sim.yearvec, results[ch], label=ch)
    ax.legend()
    ax.set_title(module)

results = sim.results.pregnancy
for ch in results.keys():
    axv[2].plot(sim.yearvec, sim.results.pregnancy[ch], label=ch)
axv[2].set_title('Pregnancy')
axv[2].legend()
fig.supylabel('Count')
fig.supxlabel('Year')
fig.tight_layout()
plt.show()


"""
Defines the People class and functions associated with making people
"""
import numpy as np
import pandas as pd
import sciris as sc
import starsim as ss
from pathlib import Path

__all__ = ['People', 'Person']


class People(sc.prettyobj):
    """
    A class to perform all the operations on the people
    This class is usually created automatically by the sim. The only required input
    argument is the population size, but typically the full parameters dictionary
    will get passed instead since it will be needed before the People object is
    initialized.

    Note that this class handles the mechanics of updating the actual people,
    as well as the additional housekeeping methods (saving, loading, exporting, etc.).

    Args:
        pars (dict): the sim parameters, e.g. sim.pars -- alternatively, if a number, interpreted as n_agents
        age_data (dataframe): a dataframe of years and population sizes, if available
        extra_states (list): non-default states to initialize

    **Examples**::
        ppl = ss.People(2000)
    """

    def __init__(self, n_agents, age_data=None, extra_states=None):
        """ Initialize """

        # We internally store states in a dict keyed by the memory ID of the state, so that we can have colliding names
        # e.g., across modules, but we will never change the size of a State multiple times in the same iteration over
        # _states. This is a hidden variable because it is internally used to synchronize the size of all States contained
        # within the sim, regardless of where they are. In contrast, `People.states` offers a more user-friendly way to access
        # a selection of the states e.g., module states could be added in there, while intervention states might not
        self._states = {}
        self.version = ss.__version__  # Store version info
        self.initialized = False

        # Handle the three fundamental arrays: UIDs for tracking agents, slots for
        # tracking random numbers, and AUIDs for tracking alive agents
        n = int(n_agents)
        uids = ss.uids(np.arange(n))
        self.auids = uids.copy() # This tracks all active UIDs (in practice, agents who are alive)
        self.uid = ss.IndexArr('uid')  # This variable tracks all UIDs
        self.slot = ss.IndexArr('slot') # A slot is a special state managed internally
        self.parent = ss.IndexArr('parent', label='UID of parent')  # UID of parent, if any, IndexArray?
        self.uid.grow(new_vals=uids)
        self.slot.grow(new_vals=uids)
        self.parent.grow(new_uids=uids, new_vals=np.full(len(uids), self.parent.nan))
        for state in [self.uid, self.slot]:
            state.people = self # Manually link to people since we don't want to link to states

        # Handle additional states
        extra_states = sc.promotetolist(extra_states)
        states = [
            ss.State('alive', default=True),  # Time index for death
            ss.State('female', default=ss.bernoulli(name='female', p=0.5)),
            ss.FloatArr('age', default=self.get_age_dist(age_data)), # NaN until conceived
            ss.FloatArr('ti_dead'),  # Time index for death
            ss.FloatArr('scale', default=1.0), # The scale factor for the agents (multiplied for making results)
        ]
        states.extend(extra_states)
        self.states = ss.ndict(type=ss.Arr)
        for state in states:
            self.states.append(state, overwrite=False)
            setattr(self, state.name, state)
            state.link_people(self)

        return

    @staticmethod
    def get_age_dist(age_data):
        """
        Return an age distribution based on provided data

        The data should be provided in the form of either an Nx2 array, a pandas series
        with age as the index and counts/probability as the value, or a pandas DataFrame
        with "age" and "value" as columns. Each of these should look like e.g.:

            age      value
            0      220.548
            1      206.188
            2      195.792
            3      187.442

        The ages will be interpreted as lower bin edges. An upper bin edge will
        automatically be added based on the final age plus the difference of the
        last two bins. To explicitly control the width of the upper age bin, add
        an extra entry to the ``age_data`` with a value of 0 and an age value
        corresponding to the desired upper age bound.

        Args:
            age_data: An array/series/dataframe with an index corresponding to age values, and a value corresponding to histogram counts
                         or relative proportions. A distribution will be estimated based on the histogram. The histogram will be
                         assumed to correspond to probability densitiy if the sum of the histogram values is equal to 1, otherwise
                         it will be assumed to correspond to counts.

        Note: age_data can also be provided as a string

        Returns:
            An ``ss.Dist`` instance that returns an age for newly created agents
        """
        if age_data is None:
            dist = ss.uniform(low=0, high=100, name='Age distribution')
        else:
            # Try loading from file
            if isinstance(age_data, str) or isinstance(age_data, Path):
                age_data = pd.read_csv(age_data)

            # Process
            if isinstance(age_data, np.ndarray): # TODO: accept output of np.histogram()
                age_bins = age_data[:,0]
                age_props = age_data[:,1]
            elif isinstance(age_data, pd.Series):
                age_bins = age_data.index
                age_props = age_data.values
            elif isinstance(age_data, pd.DataFrame):
                age_bins = age_data['age'].values
                age_props = age_data['value'].values

            # Convert to a histogram
            dist = ss.histogram(values=age_props, bins=age_bins, name='Age distribution')

        return dist

    def link_sim(self, sim):
        """ Initialization """
        if self.initialized:
            errormsg = 'Cannot re-initialize a People object directly; use sim.init(reset=True)'
            raise RuntimeError(errormsg)
        self.sim = sim # Store the sim
        ss.link_dists(obj=self.states, sim=sim, module=self, skip=[ss.Sim, ss.Module])
        return

    def add_module(self, module, force=False):
        """
        Add a Module to the People instance

        This method is used to add a module to the People. It will register any module states with this
        people instance for dynamic resizing, and expose the states contained in the module to the user
        via `People.states.<module_name>.<state_name>`

        The entries created below make it possible to do `sim.people.hiv.susceptible` or
        `sim.people.states['hiv.susceptible']` and have both of them work
        """
        # Map the module's states into the People state ndict
        if hasattr(self, module.name) and not force:
            raise Exception(f'Module {module.name} already added')

        if len(module.states):
            module_states = sc.objdict()
            setattr(self, module.name, module_states)
            for state in module.states:
                state.link_people(self)
                combined_name = module.name + '.' + state.name  # We will have to resolve how this works with multiple instances of the same module (e.g., for strains). The underlying machinery should be fine though, with People._states being flat and keyed by ID
                self.states[combined_name] = state # Register the state on the user-facing side using the combined name. Within the original module, it can still be referenced by its original name
                module_states[state.name] = state
        return

    def init_vals(self):
        """ Populate states with initial values, the final step of initialization """
        for state in self.states():
            if not state.initialized:
                state.init_vals()
        self.initialized = True
        return

    def __bool__(self):
        """ Ensure that zero-length people are still truthy """
        return True

    def __len__(self):
        """ Length of people """
        return len(self.auids)

    @property
    def n_uids(self):
        return self.uid.len_used

    def _link_state(self, state, die=True):
        """
        Link a state with the People instance for dynamic resizing; usually called by
        state.link_people()

        All states should be registered by this function for the purpose of connecting them to the
        People's UIDs and to have them be automatically resized when the number of agents changes.
        This operation is normally triggered as part of initializing the state (via `State.init()`)
        """
        if id(state) not in self._states:
            self._states[id(state)] = state
        elif die:
            errormsg = f'Cannot add state {state} since already added'
            raise ValueError(errormsg)
        return

    def grow(self, n=None, new_slots=None):
        """
        Increase the number of agents

        Args:
            n: Integer number of agents to add
            new_slots: Optionally specify the slots to assign for the new agents. Otherwise, it will default to the new UIDs
        """
        if n is None:
            if new_slots is None:
                errormsg = 'Must supply either n or new_slots'
                raise ValueError(errormsg)
            else:
                n = len(new_slots)

        if n == 0:
            return np.array([], dtype=ss.dtypes.int)

        start_uid = self.uid.len_used
        stop_uid = start_uid + n
        new_uids = ss.uids(np.arange(start_uid, stop_uid))
        self.uid.grow(new_uids, new_vals=new_uids)

        # We need to grow the slots as well
        new_slots = new_slots if new_slots is not None else new_uids
        self.slot.grow(new_uids, new_vals=new_slots)

        self.parent.grow(new_uids, new_vals=self.parent.nan) # Grow parent array

        # Grow the states
        for state in self._states.values():
            state.grow(new_uids)

        # Finally, update the alive indices
        self.auids = self.auids.concat(new_uids)
        return new_uids

    def __getitem__(self, key):
        """
        Allow people['attr'] instead of getattr(people, 'attr')
        If the key is an integer, alias `people.person()` to return a `Person` instance
        """
        if isinstance(key, int):
            return self.person(key)
        else:
            return getattr(self, key)

    def __setitem__(self, key, value):
        """ Ditto """
        return setattr(self, key, value)

    def __iter__(self):
        """ Iterate over people """
        for i in range(len(self)):
            yield self[i]

    def __setstate__(self, state):
        """
        Set the state upon unpickling/deepcopying

        If a People instance is copied (by any mechanism) then the keys in the `_states`
        registry will no longer match the memory addresses of the new copied states. Therefore,
        after copying, we need to re-create the states registry with the new object IDs
        """
        state['_states'] =  {id(v):v for v in state['_states'].values()}
        self.__dict__ = state

        return

    def scale_flows(self, inds):
        """
        Return the scaled versions of the flows -- replacement for len(inds)
        followed by scale factor multiplication
        """
        return self.scale[inds].sum()

    def update_post(self):
        """ Final updates at the very end of the timestep """
        sim = self.sim
        if sim.pars.use_aging:
            self.age[self.alive.uids] += sim.t.dt_year
        return

    def request_death(self, uids):
        """
        External-facing function to request an agent die at the current timestep

        In general, users should not directly interact with `People.ti_dead` to minimize
        interactions between modules (e.g., if a module requesting a future death, overwrites
        death due to a different module taking place at the current timestep).

        Modules that have a future time of death (e.g., due to disease duration) should keep
        track of that internally. When the module is ready to cause the agent to die, it should
        call this method, and can update its own results for the cause of death. This way, if
        multiple modules request death on the same day, they can each record a death due to their
        own cause.

        The actual deaths are resolved after modules have all run, but before analyzers. That way,
        regardless of whether removing dead agents is enabled or not, analyzers will be able to
        see and record outcomes for agents that died this timestep.

        **WARNING** - this function allows multiple modules to each independently carry out and
        record state changes associated with death. It is therefore important that they can
        guarantee that after requesting death, the death is guaranteed to occur.

        :param uids: Agent IDs to request deaths for
        :return: UIDs of agents that have been scheduled to die on this timestep
        """
        self.ti_dead[uids] = self.sim.ti
        return

    def step_die(self):
        """ Carry out any deaths that took place this timestep """
        death_uids = (self.ti_dead <= self.sim.ti).uids
        self.alive[death_uids] = False

        # Execute deaths that took place this timestep (i.e., changing the `alive` state of the agents). This is executed
        # before analyzers have run so that analyzers are able to inspect and record outcomes for agents that died this timestep
        for disease in self.sim.diseases():
            if isinstance(disease, ss.Disease):
                disease.step_die(death_uids)

        return death_uids

    def remove_dead(self):
        """
        Remove dead agents
        """
        uids = self.dead.uids
        if len(uids):

            # Remove the UIDs from the networks too
            for network in self.sim.networks.values():
                network.remove_uids(uids) # TODO: only run once every nth timestep

            # Calculate the indices to keep
            self.auids = self.auids[np.isin(self.auids, np.unique(uids), assume_unique=True, invert=True, kind='sort')]

        return

    @property
    def dead(self):
        """ Dead boolean """
        return ~self.alive

    @property
    def male(self):
        """ Male boolean """
        return ~self.female

    def update_results(self):
        ti = self.sim.ti
        res = self.sim.results
        res.n_alive[ti] = np.count_nonzero(self.alive)
        res.new_deaths[ti] = np.count_nonzero(self.ti_dead == ti)
        res.cum_deaths[ti] = np.sum(res.new_deaths[:ti]) # TODO: inefficient to compute the cumulative sum on every timestep!
        return

    def finish_step(self):
        # self.update_results() # This is called separately
        self.remove_dead()
        self.update_post()
        return

    def person(self, ind):
        """ Get all the properties for a single person """
        person = Person()
        for key in ['uid', 'slot']:
            person[key] = self[key][ind]
        for key in self.states.keys():
            person[key] = self.states[key][ind]
        return person


class Person(sc.objdict):
    """ A simple class to hold all attributes of a person """
    def to_df(self):
        """ Convert to a dataframe """
        df = sc.dataframe.from_dict(self, orient='index', columns=['value'])
        df.index.name = 'key'
        return df



"""
Vignette 04: Default interventions
The individual examples are quite different so these are being kept as separate files

Features of CK vignette:
    - interventions can be imported from other *sims
"""

import numpy as np
import sciris as sc
import stisim as ss
import hpvsim as hpv  # Eventually, we should have full compatibility between sims
import hivsim as hiv

# %% 1. Default interventions

pars = sc.objdict(
    start=1980,
    end=2040,
    dt=0.2,
)

routine_hpv_vx = hpv.routine_vx(
    prob=0.9,
    years=[2025, 2030],
    product='bivalent',
    age_range=[9, 10],
    label='routine vx'
)

art_start = 2000
art = ss.interventions.ART(
    start=art_start,
    coverage=np.linspace(0, 1, (pars.end - art_start) / pars.dt)
)

interventions = [
    routine_hpv_vx,
    art,
]

hpv = hpv.Sim()  # do we want to make this work?
hiv = hiv.Sim()

sim = ss.Sim(modules=[hpv, hiv], interventions=interventions)
sim.run()
sim.plot()


# %% 2. Custom intervention

class anc_visit(ss.Intervention):

    def __init__(self, start, end, freq=1.0, link=True):
        self.start = start
        self.end = end
        self.freq = freq
        self.link = link
        self.known_stis = ['hiv', 'gonorrhea', 'syphilis', 'chlamydia']
        self.present = []
        self.results = dict()
        return

    def initialize(self, sim):
        # Check that right modules are present
        if not 'hiv' in sim.modules:
            raise ValueError('Must have HIV to use an ANC visit intervention')
        for sti in self.known_stis:
            if sti in sim.modules:
                self.present.append(sti)
        return

    def apply(self, sim):
        if (self.start <= sim.t < self.end) and (not sim.year % self.freq):

            # Initialize results of agents who are positive
            pos = sc.objdict()

            # Then check other STIs
            for sti in self.present:
                pos[sti] = sim.modules[sti].test()

            # Optionally immediately link HIV cases to care
            if self.link:
                sim.hiv.link_to_care(pos.hiv)

            self.results[sim.year] = pos


# Run
anc = anc_visit(start=2010, end=2030)
sim = ss.Sim(modules=['hiv', 'hpv', 'gonorrhea'], interventions=anc)
sim.run()
sim.plot()




"""
Vignette 08: Custom results
"""

import stisim as sts
import starsim as ss
import matplotlib.pyplot as plt


##################
# @RomeshA
##################

# Scalar results

# How would you add an item to the simulations summary?
class MyModule(Module):
    def summary(self, sim):
        # Adds _extra_ summary items - augments/updates the auto summary computed by sim.compute_summary()
        s = super().summary(sim)
        s['new_quantity'] = <something>
        return s

# Results with dimension t

class MyModule(Module):
    def init_results(self, sim):
        Module.init_results(self, sim)
        sim.results[self.name]['n_art'] = sts.Result('n_art', self.name, sim.npts, dtype=int)

    def update_results(self, sim):
        sim.results[self.name]['n_art'] = np.count_nonzero(sim.people.alive & sim.people[self.name].on_art)

# OR

class MyAnalyzer(Analyzer):
    def initialize(self, sim):
        super().initialize()
        sim.results[self.name]['n_art'] = sts.Result('n_art', self.name, sim.npts, dtype=int)

    def update(self, sim):
        sim.results[self.name]['n_art'] = np.count_nonzero(sim.people.alive & sim.people[self.name].on_art)

# Results with other dimension e.g., age bins

class MyAnalyzer(Analyzer):
    def __init__(self, age_bin_width):
        self.bins = np.arange(0,120,age_bin_width)

    def initialize(self, sim):
        super().initialize()
        self.binned_result = np.zeros((sim.npts, len(self.bins)))

    def update(self, sim):
        self.binned_result[sim.ti,:] = np.histogram(sim.people.hiv.on_art, self.bins)




import numpy as np
import sciris as sc
import pylab as pl

# PARAMETERS
p = sc.dictobj(
    n = 10, # Number of nodes
    n_sources = 2, # Number of sources (seed infections)
    edge_prob = 0.2, # Edge probability
    trans_prob = 0.05,
    seed = 94895,
    npts = 10,
)


class Sim:
    def __init__(self, p, method='random', repseed=0):
        self.p = p
        self.repseed = repseed
        self.method = method
        self.initialize()
        self.make_network()
    
    def initialize(self):
        np.random.seed(self.p.seed+self.repseed)
        n = self.p.n
        self.uids = np.arange(n)
        self.inf = np.zeros(n, dtype=int)
        self.inf[0:self.p.n_sources] = 1
        self.xpos = np.random.rand(n)
        self.ypos = np.random.rand(n)
        self.t = 0
        self.tvec = np.arange(p.npts+1)
        self.n_infected = p.n_sources*np.ones(p.npts+1)

    def make_network(self):
        p = self.p
        n_possible = p.n**2
        n_edges = np.random.binomial(n_possible, p.edge_prob)
        p1 = np.random.choice(self.uids, size=n_edges)
        p2 = np.random.choice(self.uids, size=n_edges)
        self.src = sc.cat(p1, p2) # Bidirectional transmission
        self.trg = sc.cat(p2, p1)
        
    def get_edges(self):
        possible = (self.inf[self.src]==1) * (self.inf[self.trg]==0)
        inds = sc.findinds(possible)
        return inds
    
    def trans_rand(self, inds):
        r = np.random.rand(len(inds))
        trans = self.p.trans_prob > r
        trans_inds = inds[trans]
        inf_inds = self.trg[trans_inds]
        return inf_inds
    
    def trans_modulo(self, inds):
        r_src = np.random.rand(self.p.n*10)
        r_trg = np.random.rand(self.p.n*10)
        new = np.random.rand(self.p.n*10)
        src_inds = self.src[inds]*5
        trg_inds = self.trg[inds]
        comb_inds = (10*self.p.n-src_inds+trg_inds) % self.p.n*10
        new_inds = np.random.randint(0,self.p.n,len(inds))
        r1 = r_src[src_inds]
        r2 = r_trg[trg_inds]
        r3 = r_src[comb_inds]
        r4 = r_trg[comb_inds]
        r5 = new[comb_inds]
        r = (r1 + r2 + r3 + r4 + r5) % 1.0
        trans = self.p.trans_prob > r
        trans_inds = inds[trans]
        inf_inds = self.trg[trans_inds]
        return inf_inds
    
    def trans_roulette(self, inds):
        trans_prob = self.p.trans_prob
        probs = np.ones(self.p.n)
        for ind in inds:
            probs[self.trg[ind]] *= (1-trans_prob)
        probs = 1 - probs
        r = np.random.rand(self.p.n)
        inf_inds = probs > r
        return inf_inds
        
    def step(self):
        self.t += 1
        inds = self.get_edges()
        if self.method == 'random':
            inf_inds = self.trans_rand(inds)
        elif self.method == 'modulo':
            inf_inds = self.trans_modulo(inds)
        elif self.method == 'roulette':
            inf_inds = self.trans_roulette(inds)
        self.inf[inf_inds] = True
        self.n_infected[self.t] = self.inf.sum()
            
    def run(self):
        for t in range(self.p.npts):
            self.step()
            if self.n_infected[self.t] == self.p.n:
                self.n_infected[self.t+1:] = self.p.n
                break
        return
            
    def plot_network(self):
        fig = pl.figure()
        x = self.xpos
        y = self.ypos
        inf = self.inf == 1
        ninf = self.inf == 0
        pl.scatter(x[ninf], y[ninf], c='green')
        pl.scatter(x[inf], y[inf], c='red')
        for p1,p2 in zip(self.src, self.trg):
            pl.plot([x[p1], x[p2]], [y[p1], y[p2]], alpha=0.2, c='k', lw=0.1)
        return fig
    
    def plot(self):
        fig = pl.figure()
        pl.plot(self.tvec, self.n_infected)
        return fig


def run_sim(sim):
    sim.run()
    return sim


class MultiSim:
    def __init__(self, p, method='random', reps=10):
        self.p = p
        self.method = method
        self.reps = reps
        self.initialize()
    
    def initialize(self):
        self.sims = [Sim(p=self.p, method=self.method, repseed=rep) for rep in range(self.reps)]
    
    def run(self):
        self.sims = sc.parallelize(run_sim, self.sims, parallelizer='thread')
        self.results = np.zeros((self.p.npts+1, self.reps))
        for i,sim in enumerate(self.sims):
            self.results[:,i] = sim.n_infected
        
    def plot_all(self):
        pl.figure()
        pl.plot(self.results, alpha=0.2)
    
    def plot_stats(self):
        pl.figure()
        qs = [0, 0.0025, 0.01, 0.05, 0.1, 0.25]
        res = self.results
        mean = res.mean(axis=1)
        std = res.std(axis=1)
        x = np.arange(len(mean))
        for q in qs:
            low = np.quantile(res, q, axis=1)
            high = np.quantile(res, 1-q, axis=1)
            pl.fill_between(x, low, high, alpha=0.1, facecolor='red')
        pl.plot(mean, 'o-', c='k', lw=2)
        pl.title(f'Method: {self.method}; stats: μ={mean.mean():0.4f}, σ={std.mean():0.4f}')
        
        
if __name__ == '__main__':
    
    sc.options(dpi=200)
    T = sc.timer()
    for method in ['random', 'modulo', 'roulette']:
        T.tic()
        msim = MultiSim(p=p, method=method, reps=1_000)
        msim.run()
        T.toc(method)
        msim.plot_stats()
    pl.show()


"""
Define array-handling classes, including agent states
"""
import numpy as np
import sciris as sc
import starsim as ss

# Shorten these for performance
ss_float = ss.dtypes.float
ss_int   = ss.dtypes.int
ss_bool  = ss.dtypes.bool
type_def = {
    ss_float: ('float', float, np.float64, np.float32),
    ss_int: ('int', int, np.int64, np.int32),
    ss_bool: ('bool', bool, np.bool_),
}
type_map = {v:k for k,vlist in type_def.items() for v in vlist} # Invert into a full dictionary

__all__ = ['BaseArr', 'Arr', 'FloatArr', 'BoolArr', 'State', 'IndexArr', 'uids']


class BaseArr(np.lib.mixins.NDArrayOperatorsMixin):
    """
    An object that acts exactly like a NumPy array, except stores the values in self.values.
    """
    def __init__(self, values, *args, **kwargs):
        self.values = np.array(values, *args, **kwargs)
        return

    def __getattr__(self, attr):
        """ Make it behave like a regular array mostly -- enables things like sum(), mean(), etc. """
        if attr in ['__deepcopy__', '__getstate__', '__setstate__']:
            return self.__getattribute__(attr)
        else:
            return object.__getattribute__(self, 'values').__getattribute__(attr) # Be explicit to avoid possible recurison

    # Define more base methods
    def __len__(self):   return self.values.__len__()
    def __bool__(self):  return self.values.__bool__()
    def __int__(self):   return self.values.__int__()
    def __float__(self): return self.values.__float__()
    def __contains__(self, key): return self.values.__contains__(key)

    def convert(self, obj):
        """ Check if an object is an array, and convert if so """
        if isinstance(obj, np.ndarray):
            return self.asnew(obj)
        elif isinstance(obj, BaseArr):
            return self.asnew(obj.values)
        return obj

    @staticmethod
    def _arr(obj):
        """ Helper function to efficiently extract values from a BaseArr, or return the original object """
        if isinstance(obj, BaseArr):
            return obj.values
        return obj

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        """ To handle all numpy operations, e.g. arr1*arr2 """
        # Convert all inputs to their .values if they are BaseArr, otherwise leave unchanged
        inputs = [self._arr(x) for x in inputs]
        kwargs = {k:self._arr(v) for k,v in kwargs.items()}
        result = getattr(ufunc, method)(*inputs, **kwargs)

        # If result is a tuple (e.g., for divmod or ufuncs that return multiple values), convert all results to BaseArr
        if isinstance(result, tuple):
            return tuple(self.convert(x) for x in result)

        result = self.convert(result)
        return result

    def __iter__(self):
        """ For iterating correctly, e.g. for sum() """
        return iter(self.values)

    def __getitem__(self, index):
        """ For indexing and slicing, e.g. arr[inds] """
        return self.values[index]

    def __setitem__(self, index, value):
        """ Assign values, e.g. arr1[inds] = arr2 """
        self.values[index] = value

    def __array__(self, *args, **kwargs):
        """ To ensure isinstance(arr, BaseArr) passes when creating new instances """
        return np.array(self.values, *args, **kwargs)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.values})"

    def disp(self):
        """ Full display of object """
        return sc.pr(self)

    def asnew(self, values=None, cls=None, **kwargs):
        """ Duplicate and copy (rather than link) data """
        if cls is None: # Use the current class if none is provided
            cls = self.__class__
        new = object.__new__(cls) # Create a new Arr instance
        new.__dict__ = self.__dict__.copy() # Copy pointers
        new.__dict__.update(kwargs) # Update any keyword arguments provided # TODO: add validation?
        if values is not None:
            new.values = values # Replace data with new data
        else:
            new.values = sc.cp(new.values) # TODO: is this needed?
        return new

    def update(self, skip_none=True, overwrite=True, force=False, **kwargs):
        """ Update the attributes, skipping None values and raising an error if extra keys are added """
        if skip_none: # Remove None inputs
            kwargs = {k:v for k,v in kwargs.items() if v is not None}
        if not overwrite: # Don't overwrite non-None values
            kwargs = {k:v for k,v in kwargs.items() if getattr(self, k, None) is None}
        if not force: # Check if we'd be creating any new keys
            kw_keys = set(kwargs.keys())
            self_keys = set(self.__dict__.keys())
            diff = kw_keys - self_keys
            if diff:
                errormsg = f'Invalid arguments to {self}: {diff}'
                raise ValueError(errormsg)
        self.__dict__.update(kwargs) # Actually perform the update
        return self

    def to_json(self):
        """ Return a dictionary representation of the Arr """
        out = dict(
            classname = self.__class__.__name__,
            values = sc.jsonify(self.values),
        )
        return out


class Arr(BaseArr):
    """
    Store a state of the agents (e.g. age, infection status, etc.) as an array.

    In practice, ``Arr`` objects can be used interchangeably with NumPy arrays.
    They have two main data interfaces: ``Arr.raw`` contains the "raw", underlying
    NumPy array of the data. ``Arr.values`` contains the "active" values, which
    usually corresponds to agents who are alive.

    By default, operations are performed on active agents only (specified by ``Arr.auids``,
    which is a pointer to ``sim.people.auids``). For example, ``sim.people.age.mean()``
    will only use the ages of active agents. Thus, ``sim.people.age.mean()``
    is equal to ``sim.people.age.values.mean()``, not ``sim.people.age.raw.mean()``.

    If indexing by an int or slice, ``Arr.values`` is used. If indexing by an
    ``ss.uids`` object, ``Arr.raw`` is used. ``Arr`` objects can't be directly
    indexed by a list or array of ints, as this would be ambiguous about whether
    ``values`` or ``raw`` is intended. For example, if there are 1000 people in a
    simulation and 100 of them have died, ``sim.people.age[999]`` will return
    an ``IndexError`` (since ``sim.people.age[899]`` is the last active agent),
    whereas ``sim.people.age[ss.uids(999)]`` is valid.

    Args:
        name (str): The name for the state (also used as the dictionary key, so should not have spaces etc.)
        dtype (class): The dtype to use for this instance (if None, infer from value)
        default (any): Specify default value for new agents. This can be
        - A scalar with the same dtype (or castable to the same dtype) as the State
        - A callable, with a single argument for the number of values to produce
        - A ``ss.Dist`` instance
        nan (any): the value to use to represent NaN (not a number); also used as the default value if not supplied
        label (str): The human-readable name for the state
        skip_init (bool): Whether to skip initialization with the People object (used for uid and slot states)
        people (ss.People): Optionally specify an initialized People object, used to construct temporary Arr instances
    """
    def __init__(self, name=None, dtype=None, default=None, nan=None, label=None, skip_init=False, people=None):
        # Set attributes
        self.name = name
        self.label = label or name
        self.default = default
        self.nan = nan
        self.dtype = dtype
        self.people = people # Used solely for accessing people.auids

        if self.people is None:
            # This Arr is being defined in advance (e.g., as a module state) and we want a bidirectional link
            # with a People instance for dynamic growth. These properties will be initialized later when the
            # People/Sim are initialized
            self.len_used = 0
            self.len_tot = 0
            self.initialized = skip_init
            self.raw = np.empty(0, dtype=dtype)
        else:
            # This Arr is a temporary object used for intermediate calculations when we want to index an array
            # by UID (e.g., inside an update() method). We allow this state to reference an existing, initialized
            # People object, but do not register it for dynamic growth
            self.len_used = self.people.uid.len_used
            self.len_tot = self.people.uid.len_tot
            self.initialized = True
            self.raw = np.full(self.len_tot, dtype=self.dtype, fill_value=self.nan)

        return

    def __repr__(self):
        arr_str = np.array2string(self.values, max_line_width=200)
        if self.name:
            string = f'<{self.__class__.__name__} "{str(self.name)}", len={len(self)}, {arr_str}>'
        else:
            string = f'<{self.__class__.__name__}, len={len(self)}, {arr_str}>'
        return string

    def __len__(self):
        return len(self.auids)

    def _convert_key(self, key):
        """
        Used for getitem and setitem to determine whether the key is indexing
        the raw array (``raw``) or the active agents (``values``), and to convert
        the key to array indices if needed.
        """
        if isinstance(key, (uids, int, ss_int)):
            return key
        elif isinstance(key, (BoolArr, IndexArr)):
            return key.uids
        elif isinstance(key, slice):
            return self.auids[key]
        elif not np.isscalar(key) and len(key) == 0: # Handle [], np.array([]), etc.
            return uids()
        else:
            errormsg = f'Indexing an Arr ({self.name}) by ({key}) is ambiguous or not supported. Use ss.uids() instead, or index Arr.raw or Arr.values.'
            raise Exception(errormsg)

    def __getitem__(self, key):
        key = self._convert_key(key)
        return self.raw[key]

    def __setitem__(self, key, value):
        key = self._convert_key(key)
        self.raw[key] = value
        return

    def __gt__(self, other): return self.asnew(self.values > other,  cls=BoolArr)
    def __lt__(self, other): return self.asnew(self.values < other,  cls=BoolArr)
    def __ge__(self, other): return self.asnew(self.values >= other, cls=BoolArr)
    def __le__(self, other): return self.asnew(self.values <= other, cls=BoolArr)
    def __eq__(self, other): return self.asnew(self.values == other, cls=BoolArr)
    def __ne__(self, other): return self.asnew(self.values != other, cls=BoolArr)

    def __and__(self, other): raise BooleanOperationError(self)
    def __or__(self, other):  raise BooleanOperationError(self)
    def __xor__(self, other): raise BooleanOperationError(self)
    def __invert__(self):     raise BooleanOperationError(self)

    @property
    def auids(self):
        """ Link to the indices of active agents -- sim.people.auids """
        try:
            return self.people.auids
        except:
            if not self.initialized:
                ss.warn('Trying to access non-initialized Arr object; in most cases, Arr objects need to be initialized with a Sim object, but set skip_init=True if this is intentional.')
            return uids(np.arange(len(self.raw)))

    def count(self):
        return np.count_nonzero(self.values)

    @property
    def values(self):
        """ Return the values of the active agents """
        return self.raw[self.auids]

    def set(self, uids, new_vals=None):
        """ Set the values for the specified UIDs"""
        if new_vals is None:
            if isinstance(self.default, ss.Dist):
                new_vals = self.default.rvs(uids)
            elif callable(self.default):
                new_vals = self.default(len(uids))
            elif self.default is not None:
                new_vals = self.default
            else:
                new_vals = self.nan
        self.raw[uids] = new_vals
        return new_vals

    def set_nan(self, uids):
        """ Shortcut function to set values to NaN """
        self.raw[uids] = self.nan
        return

    @property
    def isnan(self):
        return self.asnew(self.values == self.nan, cls=BoolArr)

    @property
    def notnan(self):
        return self.asnew(self.values != self.nan, cls=BoolArr)

    def grow(self, new_uids=None, new_vals=None):
        """
        Add new agents to an Arr

        This method is normally only called via `People.grow()`.

        Args:
            new_uids: Numpy array of UIDs for the new agents being added
            new_vals: If provided, assign these state values to the new UIDs
        """
        orig_len = self.len_used
        n_new = len(new_uids)
        self.len_used += n_new  # Increase the count of the number of agents by `n` (the requested number of new agents)

        # Physically reshape the arrays, if needed
        if orig_len + n_new > self.len_tot:
            n_grow = max(n_new, self.len_tot//2)  # Minimum 50% growth, since growing arrays is slow
            new_empty = np.empty(n_grow, dtype=self.dtype) # 10x faster than np.zeros()
            self.raw = np.concatenate([self.raw, new_empty], axis=0)
            self.len_tot = len(self.raw)
            if n_grow > n_new: # We added extra space at the end, set to NaN
                nan_uids = np.arange(self.len_used, self.len_tot)
                self.set_nan(nan_uids)

        # Set new values, and NaN if needed
        self.set(new_uids, new_vals=new_vals) # Assign new default values to those agents
        return

    def link_people(self, people):
        """ Link a People object to this state, for access auids """
        self.people = people # Link the people object to this state
        people._link_state(self) # Ensure the state is linked to the People object as well
        return

    def init_vals(self):
        """ Actually populate the initial values and mark as initialized; only to be used on initialization """
        if self.initialized:
            errormsg = f'Cannot re-initialize state {self}; use set() instead'
            raise RuntimeError(errormsg)
        self.grow(self.people.uid)
        self.initialized = True
        return

    def asnew(self, arr=None, cls=None, name=None):
        """ Duplicate and copy (rather than link) data, optionally resetting the array """
        if cls is None:
            cls = self.__class__
        if arr is None:
            arr = self.values
        new = object.__new__(cls) # Create a new Arr instance
        new.__dict__ = self.__dict__.copy() # Copy pointers
        new.dtype = arr.dtype # Set to correct dtype
        new.name = name # In most cases, the asnew Arr has different values to the original Arr so the original name no longer makes sense
        new.raw = np.empty(new.raw.shape, dtype=new.dtype) # Copy values, breaking reference
        new.raw[new.auids] = arr
        return new

    def true(self):
        """ Efficiently convert truthy values to UIDs """
        return self.auids[self.values.astype(bool)]

    def false(self):
        """ Reverse of true(); return UIDs of falsy values """
        return self.auids[~self.values.astype(bool)]

    def to_json(self):
        """ Export to JSON """
        out = dict(
            classname = self.__class__.__name__,
            name = self.name,
            label = self.label,
            default = self.default,
            nan = self.nan,
            dtype = self.dtype,
            values = sc.jsonify(self.values),
        )
        return out


class FloatArr(Arr):
    """
    Subclass of Arr with defaults for floats and ints.

    Note: Starsim does not support integer arrays by default since they introduce
    ambiguity in dealing with NaNs, and float arrays are suitable for most purposes.
    If you really want an integer array, you can use the default Arr class instead.
    """
    def __init__(self, name=None, nan=np.nan, **kwargs):
        super().__init__(name=name, dtype=ss_float, nan=nan, **kwargs)
        return

    @property
    def isnan(self):
        """ Return BoolArr for NaN values """
        return self.asnew(np.isnan(self.values), cls=BoolArr)

    @property
    def notnan(self):
        """ Return BoolArr for non-NaN values """
        return self.asnew(~np.isnan(self.values), cls=BoolArr)

    @property
    def notnanvals(self):
        """ Return values that are not-NaN """
        vals = self.values # Shorten and avoid double indexing
        out = vals[np.nonzero(~np.isnan(vals))[0]]
        return out


class BoolArr(Arr):
    """ Subclass of Arr with defaults for booleans """
    def __init__(self, name=None, nan=False, **kwargs): # No good NaN equivalent for bool arrays
        super().__init__(name=name, dtype=ss_bool, nan=nan, **kwargs)
        return

    def __and__(self, other): return self.asnew(self.values & other)
    def __or__(self, other):  return self.asnew(self.values | other)
    def __xor__(self, other): return self.asnew(self.values ^ other)
    def __invert__(self):     return self.asnew(~self.values)

    # BoolArr cannot store NaNs so report all entries as being not-NaN
    @property
    def isnan(self):
        return self.asnew(np.full_like(self.values, fill_value=False), cls=BoolArr)

    @property
    def notnan(self):
        return self.asnew(np.full_like(self.values, fill_value=True), cls=BoolArr)

    @property
    def uids(self):
        """ Alias to Arr.true """
        return self.true()

    def split(self):
        """ Return UIDs of values that are true and false as separate arrays """
        t_uids = self.true()
        f_uids = self.false()
        return t_uids, f_uids


class State(BoolArr):
    """
    A boolean array being used as a state.

    Although functionally identical to BoolArr, a State is handled differently in
    terms of automation: specifically, results are automatically generated from a
    State (but not a BoolArr).

    States are typically used to keep track of externally-facing variables (e.g.
    disease.susceptible), while BoolArrs can be used to keep track of internal
    ones (e.g. disease.has_immunity).
    """
    pass


class IndexArr(Arr):
    """ A special class of Arr used for UIDs and RNG IDs; not to be used as an integer array (for that, use FloatArr) """
    def __init__(self, name=None, label=None):
        super().__init__(name=name, dtype=ss_int, default=None, nan=-1, label=label, skip_init=True)
        self.raw = uids(self.raw)
        return

    @property
    def uids(self):
        """ Alias to self.values, to allow Arr.uids like BoolArr """
        return self.values

    def grow(self, new_uids=None, new_vals=None):
        """ Change the size of the array """
        if new_uids is None and new_vals is not None: # Used as a shortcut to avoid needing to supply twice
            new_uids = new_vals
        super().grow(new_uids=new_uids, new_vals=new_vals)
        self.raw = uids(self.raw)
        return


class uids(np.ndarray):
    """
    Class to specify that integers should be interpreted as UIDs.

    For all practical purposes, behaves like a NumPy integer array. However,
    has additional methods ``uids.concat()`` (instance method), ``ss.uids.cat()``
    (class method), ``uids.remove()``, and ``uids.intersect()`` to simplify common
    UID operations.
    """
    def __new__(cls, arr=None):
        if isinstance(arr, np.ndarray): # Shortcut to typical use case, where the input is an array
            return arr.astype(ss_int).view(cls)
        elif isinstance(arr, BoolArr): # Shortcut for arr.uids
            return arr.uids
        elif isinstance(arr, set):
            return np.fromiter(arr, dtype=ss_int).view(cls)
        elif arr is None: # Shortcut to return empty
            return np.empty(0, dtype=ss_int).view(cls)
        elif isinstance(arr, int): # Convert e.g. ss.uids(0) to ss.uids([0])
            arr = [arr]
        return np.asarray(arr, dtype=ss_int).view(cls) # Handle everything else

    def concat(self, other, **kw): # Class and instance methods can't share a name
        """ Equivalent to np.concatenate(), but return correct type """
        return np.concatenate([self, other], **kw).view(self.__class__)

    @classmethod
    def cat(cls, *args, **kw):
        """ Equivalent to np.concatenate(), but return correct type """
        if len(args) == 0 or (len(args) == 1 and (args[0] is None or not len(args[0]))):
            return uids()
        arrs = args[0] if len(args) == 1 else args # TODO: handle one-array case
        return np.concatenate(arrs, **kw).view(cls)

    def remove(self, other, **kw):
        """ Remove provided UIDs from current array"""
        if isinstance(other, BoolArr):
            other = other.uids
        return np.setdiff1d(self, other, **kw).view(self.__class__)

    def intersect(self, other, **kw):
        """ Keep only UIDs that are also present in the other array """
        if isinstance(other, BoolArr):
            other = other.uids
        return np.intersect1d(self, other, **kw).view(self.__class__)

    def union(self, other, **kw):
        """ Return all UIDs present in both arrays """
        if isinstance(other, BoolArr):
            other = other.uids
        return np.union1d(self, other, **kw).view(self.__class__)

    def xor(self, other, **kw):
        """ Return UIDs present in only one of the arrays """
        if isinstance(other, BoolArr):
            other = other.uids
        return np.setxor1d(self, other, **kw).view(self.__class__)

    def to_numpy(self):
        """ Return a view as a standard NumPy array """
        return self.view(np.ndarray)

    def unique(self, return_index=False):
        """ Return unique UIDs; equivalent to np.unique() """
        if return_index:
            arr, index = np.unique(self, return_index=True)
            return arr.view(self.__class__), index
        else:
            arr = np.unique(self).view(self.__class__)
            return arr

    # Implement collection of operators
    def __and__(self, other): return self.intersect(other)
    def __or__(self, other) : return self.union(other)
    def __sub__(self, other): return self.remove(other)
    def __xor__(self, other): return self.xor(other)
    def __invert__(self)    : raise Exception(f"Cannot invert an instance of {self.__class__.__name__}. One possible cause is attempting `~x.uids` - use `x.false()` or `(~x).uids` instead")


class BooleanOperationError(NotImplementedError):
    """ Raised when a logical operation is performed on a non-logical array """
    def __init__(self, arr):
        msg = f'Logical operations are only valid on Boolean arrays, not {arr.dtype}'
        super().__init__(msg)



import starsim as ss
import pandas as pd

# Read in age-specific fertility rates
fertility_rates = pd.read_csv('/home/cliffk/idm/starsim/docs/tutorials/test_data/nigeria_asfr.csv')
pregnancy = ss.Pregnancy(pars={'fertility_rate': fertility_rates})

death_rates = pd.read_csv('/home/cliffk/idm/starsim/docs/tutorials/test_data/nigeria_deaths.csv')
death = ss.Deaths(pars={'death_rate': death_rates, 'units': 1})

demographics = [pregnancy, death]

# Make people using the distribution of the population by age/sex in 1995
n_agents = 5_000
nga_pop_1995 = 106819805  # Population of Nigeria in 1995, the year we will start the model
age_data = pd.read_csv('/home/cliffk/idm/starsim/docs/tutorials/test_data/nigeria_age.csv')
ppl = ss.People(n_agents, age_data=age_data)

# Make the sim, run and plot
sim = ss.Sim(total_pop=nga_pop_1995, start=1995, people=ppl, demographics=demographics, networks='random', diseases='sir')
sim.run()

# Read in a file with the actual population size
nigeria_popsize = pd.read_csv('/home/cliffk/idm/starsim/docs/tutorials/test_data/nigeria_popsize.csv')
data = nigeria_popsize[(nigeria_popsize.year >= 1995) & (nigeria_popsize.year <= 2030)]

# Plot the overall population size - simulated vs data
import matplotlib.pyplot as plt
fig, ax = plt.subplots(1, 1)
ax.scatter(data.year, data.n_alive, alpha=0.5, label='Data')
ax.plot(sim.yearvec, sim.results.n_alive, color='k', label='Model')
ax.legend()
ax.set_title('Population')
plt.show();



import numpy as np
import sciris as sc
import pylab as pl
# import numba as nb

# @nb.njit # No faster -- not sure why not
def get_probs(n, inds, trg, trans_prob):
    probs = np.ones(n)
    trg_inds = trg[inds]
    for trg_ind in trg_inds:
        probs[trg_ind] *= (1-trans_prob)
    probs = 1 - probs
    return probs

# PARAMETERS
p = sc.objdict(
    n = 100, # Number of nodes
    n_sources = 2, # Number of sources (seed infections)
    edge_prob = 0.1, # Edge probability
    trans_prob = 0.05,
    seed = 6,
    npts = 10,
)


class Sim:
    def __init__(self, p, method='random', repseed=0):
        self.p = p
        self.repseed = repseed
        self.method = method
        self.initialize()
        self.make_network()
    
    def initialize(self):
        np.random.seed(self.p.seed+self.repseed)
        n = self.p.n
        self.uids = np.arange(n)
        self.inf = np.zeros(n, dtype=int)
        self.inf[0:self.p.n_sources] = 1
        self.xpos = np.random.rand(n)
        self.ypos = np.random.rand(n)
        self.t = 0
        self.tvec = np.arange(p.npts+1)
        self.n_infected = p.n_sources*np.ones(p.npts+1)

    def make_network(self):
        p = self.p
        n_possible = p.n**2
        n_edges = np.random.binomial(n_possible, p.edge_prob)
        p1 = np.random.choice(self.uids, size=n_edges)
        p2 = np.random.choice(self.uids, size=n_edges)
        self.src = sc.cat(p1, p2) # Bidirectional transmission
        self.trg = sc.cat(p2, p1)
        
    def get_edges(self):
        possible = (self.inf[self.src]==1) * (self.inf[self.trg]==0)
        inds = sc.findinds(possible)
        return inds
    
    def trans_rand(self, inds):
        r = np.random.rand(len(inds))
        trans = self.p.trans_prob > r
        trans_inds = inds[trans]
        inf_inds = self.trg[trans_inds]
        return inf_inds
    
    def trans_modulo(self, inds):
        r1 = np.random.rand(self.p.n)
        r2 = np.random.rand(self.p.n)
        # np.random.shuffle(r2)
        r1 = r1[self.src[inds]]
        r2 = r2[self.trg[inds]]
        r = (r1 + r2) % 1.0
        trans = self.p.trans_prob > r
        trans_inds = inds[trans]
        inf_inds = self.trg[trans_inds]
        return inf_inds
    
    def trans_roulette(self, inds):
        trans_prob = self.p.trans_prob
        n = self.p.n
        trg = self.trg
        probs = get_probs(n, inds, trg, trans_prob)
        r = np.random.rand(self.p.n)
        inf_inds = probs > r
        return inf_inds
        
    def step(self):
        self.t += 1
        inds = self.get_edges()
        if self.method == 'random':
            inf_inds = self.trans_rand(inds)
        elif self.method == 'roulette':
            inf_inds = self.trans_roulette(inds)
        elif self.method == 'modulo':
            inf_inds = self.trans_modulo(inds)
        else:
            raise Exception('Method not found')
        self.inf[inf_inds] = True
        self.n_infected[self.t] = self.inf.sum()
            
    def run(self):
        for t in range(self.p.npts):
            self.step()
            if self.n_infected[self.t] == self.p.n:
                self.n_infected[self.t+1:] = self.p.n
                break
        return
            
    def plot_network(self):
        fig = pl.figure()
        x = self.xpos
        y = self.ypos
        inf = self.inf == 1
        ninf = self.inf == 0
        pl.scatter(x[ninf], y[ninf], c='green')
        pl.scatter(x[inf], y[inf], c='red')
        for p1,p2 in zip(self.src, self.trg):
            pl.plot([x[p1], x[p2]], [y[p1], y[p2]], alpha=0.2, c='k', lw=0.1)
        return fig
    
    def plot(self):
        fig = pl.figure()
        pl.plot(self.tvec, self.n_infected)
        return fig


def run_sim(sim):
    sim.run()
    return sim


class MultiSim:
    def __init__(self, p, method='random', reps=10):
        self.p = p
        self.method = method
        self.reps = reps
        self.initialize()
    
    def initialize(self):
        self.sims = [Sim(p=self.p, method=self.method, repseed=rep) for rep in range(self.reps)]
    
    def run(self):
        self.sims = sc.parallelize(run_sim, self.sims)
        self.results = np.zeros((self.p.npts+1, self.reps))
        for i,sim in enumerate(self.sims):
            self.results[:,i] = sim.n_infected
        
    def plot_all(self):
        pl.figure()
        pl.plot(self.results, alpha=0.2)
    
    def plot_stats(self, ax=None):
        if ax is None:
            fig,ax = pl.subplots(1)
        pl.sca(ax)
        qs = [0, 0.0025, 0.01, 0.05, 0.1, 0.25]
        res = self.results
        mean = res.mean(axis=1)
        std = res.std(axis=1)
        x = np.arange(len(mean))
        for q in qs:
            low = np.quantile(res, q, axis=1)
            high = np.quantile(res, 1-q, axis=1)
            pl.fill_between(x, low, high, alpha=0.1, facecolor='red')
        pl.plot(mean, 'o-', c='k', lw=2)
        pl.xlabel('Timestep')
        pl.ylabel('Number infected')
        pl.title(f'Method: {self.method}; stats: μ={mean.mean():0.4f}, σ={std.mean():0.4f}')
        pl.xlim(left=0, right=p.npts)
        pl.ylim(bottom=0, top=p.n)
        
        
if __name__ == '__main__':
    sc.options(dpi=200)
    T = sc.timer()
    fig, axes = pl.subplots(1,3, figsize=(18,8))
    for m,method in enumerate(['random', 'roulette', 'modulo']):
        msim = MultiSim(p=p, method=method, reps=1_000)
        T.tic()
        msim.run()
        T.toc(method)
        msim.plot_stats(ax=axes[m])
    sc.figlayout()
    pl.show()


"""
Test different formats for demographic data
"""

# %% Imports and settings
import starsim as ss
import pandas as pd
import matplotlib.pyplot as plt
import numpy as np

do_plot = True


def test_fixed_death_rate():
    """ Simple fixed death rate for all agents """
    ppl = ss.People(1000)
    bdm1 = ss.Deaths(pars={'death_rate': 0.015})
    sim1 = ss.Sim(people=ppl, demographics=bdm1, label='Constant death rate')
    sim1.run()
    return sim1


def test_series_death_rate():
    """ Death rate from a pandas series """
    series_death = pd.Series(
        index=[0, 10, 20, 25, 30, 35, 40, 45, 50, 55, 60, 65, 70, 75, 80, 85, 90, 95],
        data=[0.0046355, 0.000776, 0.0014232, 0.0016693, 0.0021449, 0.0028822, 0.0039143, 0.0053676, 0.0082756, 0.01,
              0.02, 0.03, 0.04, 0.06, 0.11, 0.15, 0.21, 0.30],
    )
    ppl = ss.People(1000)
    death_rate = ss.standardize_data(data=series_death, metadata=ss.omerge({
        'data_cols': {'year': 'Time', 'sex': 'Sex', 'age': 'AgeGrpStart', 'value': 'mx'},
        'sex_keys': {'Female':'f', 'Male':'m'},
    }))
    bdm2 = ss.Deaths(pars={'death_rate': death_rate})
    sim2 = ss.Sim(people=ppl, demographics=bdm2, label='Using age-specific data from a pandas series')
    sim2.run()
    return sim2


def test_file_death_rate():
    """ Realistic death rates in dataframe format, read from a csv file """
    realistic_death = pd.read_csv(ss.root / 'tests/test_data/nigeria_deaths.csv')
    ppl = ss.People(1000)
    bdm3 = ss.Deaths(pars={'death_rate': realistic_death})
    sim3 = ss.Sim(people=ppl, demographics=bdm3, label='Realistic death rates read from a CSV file')
    sim3.run()
    return sim3


def test_file_birth_data():
    """ Test births using CSV data """
    ppl = ss.People(1000)
    realistic_birth = pd.read_csv(ss.root / 'tests/test_data/nigeria_births.csv')
    births = ss.Births(pars={'birth_rate': realistic_birth})
    sim1 = ss.Sim(people=ppl, demographics=births, label='UN birth rates read from a CSV file')
    sim1.run()
    return sim1


def test_crude_birth_data():
    """ Test births using a crude rate """
    ppl = ss.People(1000)
    births = ss.Births(pars={'birth_rate': 36, 'units': 1 / 1000})
    sim2 = ss.Sim(people=ppl, demographics=births, label='Overall crude birth rate')
    sim2.run()
    return sim2


def test_fertility_data():
    """ Testing fertility data can be added in multiple formats """
    fertility_rates = pd.read_csv(ss.root / 'tests/test_data/nigeria_asfr.csv')
    pregnancy = ss.Pregnancy(pars={'fertility_rate': fertility_rates})
    ppl = ss.People(1000)
    sim = ss.Sim(people=ppl, demographics=pregnancy, label='UN fertility rates read from a CSV file')
    sim.run()

    assert np.array_equal(sim.results.pregnancy.pregnancies, sim.results.pregnancy.births)

    return sim


if __name__ == '__main__':

    # Deaths
    sim_death1 = test_fixed_death_rate()
    sim_death2 = test_series_death_rate()
    sim_death3 = test_file_death_rate()

    # Test births
    sim_birth1 = test_file_birth_data()
    sim_birth2 = test_crude_birth_data()

    # Test fertility
    sim_fert = test_fertility_data()


    if do_plot:
        # Plot deaths
        fig, ax = plt.subplots(2, 1)
        for sim in [sim_death1, sim_death2, sim_death3]:
            ax[0].plot(sim.tivec, sim.results.deaths.new, label=sim.label)
            ax[1].plot(sim.tivec, sim.results.n_alive)

        ax[0].set_title('New background deaths')
        ax[1].set_title('Population size')
        ax[1].set_xlabel('Time step')
        ax[0].set_ylabel('Count')
        ax[1].set_ylabel('Count')
        ax[0].legend()
        fig.tight_layout()
        plt.show()

        # Plot births
        fig, ax = plt.subplots(2, 1)
        for sim in [sim_birth1, sim_birth2,]:
            ax[0].plot(sim.tivec, sim.results.births.new, label=sim.label)
            ax[1].plot(sim.tivec, sim.results.n_alive)

        ax[0].set_title('New births')
        ax[1].set_title('Population size')
        ax[1].set_xlabel('Time step')
        ax[0].set_ylabel('Count')
        ax[1].set_ylabel('Count')
        ax[0].legend()
        fig.tight_layout()
        plt.show()

        # Plot fert
        sim = sim_fert
        fig, ax = plt.subplots(2, 1)
        ax[0].plot(sim.yearvec, sim.results.pregnancy.pregnancies, label='Pregnancies')
        ax[0].plot(sim.yearvec, sim.results.pregnancy.births, ':', label='Births')
        ax[1].plot(sim.yearvec, sim.results.n_alive, label='Population')
        ax[0].set_title('Pregnancies and births')
        ax[1].set_title('Population size')
        ax[1].set_xlabel('Year')
        ax[0].set_ylabel('Count')
        ax[1].set_ylabel('Count')
        ax[0].legend()
        fig.tight_layout()

        plt.show()



"""
General module class -- base class for diseases, interventions, etc. Also
defines Analyzers and Connectors.
"""
import sciris as sc
import starsim as ss
from functools import partial

__all__ = ['module_map', 'find_modules', 'Base', 'Module', 'Analyzer', 'Connector']

module_args = ['name', 'label'] # Define allowable module arguments


def module_map(key=None):
    """ Define the mapping between module names and types """
    module_map = sc.objdict(
        networks      = ss.Network,
        demographics  = ss.Demographics,
        diseases      = ss.Disease,
        interventions = ss.Intervention,
        analyzers     = ss.Analyzer,
        connectors    = ss.Connector,
    )
    return module_map if key is None else module_map[key]


def find_modules(key=None, flat=False):
    """ Find all subclasses of Module present in Starsim, divided by type """
    modules = sc.objdict()
    modmap = module_map()
    attrs = dir(ss) # Find all attributes in Starsim (note: does not parse user code)
    for modkey, modtype in modmap.items(): # Loop over each module type
        modules[modkey] = sc.objdict()
        for attr in attrs: # Loop over each attribute (inefficient, but doesn't need to be optimized)
            item = getattr(ss, attr)
            try:
                assert issubclass(item, modtype) # Check that it's a class, and instance of this module
                low_attr = attr.lower()
                modules[modkey][low_attr] = item # It passes, so assign it to the dict
                if modkey == 'networks' and low_attr.endswith('net'): # Also allow networks without 'net' suffix
                    modules[modkey][low_attr.removesuffix('net')] = item
            except:
                pass
    if flat:
        modules = sc.objdict({k:v for vv in modules.values() for k,v in vv.items()}) # Unpack the nested dict into a flat one
    return modules if key is None else modules[key]


class Base(sc.quickobj):
    """
    The parent class for Sim and Module objects
    """
    def __bool__(self):
        """ Ensure that zero-length modules (e.g. networks) are still truthy """
        return True

    def __len__(self):
        """ The length of a module is the number of timepoints; see also len(sim) """
        try:    return self.t.npts
        except: return 0

    def disp(self, output=False, **kwargs):
        """ Display the full object """
        out = sc.prepr(self, **kwargs)
        if not output:
            print(out)
        else:
            return out

    @property
    def ti(self):
        """ Get the current module timestep """
        try:    return self.t.ti
        except: return None

    @property
    def now(self):
        """ Shortcut to self.t.now() """
        try:    return self.t.now()
        except: return None

    @property
    def timevec(self):
        """ Shortcut to self.t.timevec """
        try:    return self.t.timevec
        except: return None



class Module(Base):
    """
    The main base class for all Starsim modules: diseases, networks, interventions, etc.

    Args:
        name (str): a short, key-like name for the module (e.g. "randomnet")
        label (str): the full, human-readable name for the module (e.g. "Random network")
        kwargs (dict): passed to ss.Time() (e.g. start, stop, unit, dt)
    """

    def __init__(self, name=None, label=None, **kwargs):
        # Handle parameters
        self.pars = ss.Pars() # Usually populated via self.define_pars()
        self.set_metadata(name, label) # Usually reset as part of self.update_pars()
        self.t = ss.Time(**kwargs, name=self.name, init=False)

        # Properties to be added by init_pre()
        self.sim = None
        self.dists = None # Turned into a Dists object by sim.init_dists() if this module has dists
        self.results = ss.Results(self.name)

        # Finish initialization
        self.pre_initialized = False
        self.initialized = False
        self.finalized = False
        return

    def __call__(self, *args, **kwargs):
        """ Allow modules to be called like functions """
        return self.step(*args, **kwargs)

    def _reconcile(self, key, value=None, default=None):
        """ Reconcile module attributes, parameters, and input arguments """
        parval = self.pars.get(key)
        attrval = getattr(self, key, parval)
        val = sc.ifelse(value, attrval, default)
        return val

    def set_metadata(self, name=None, label=None):
        """ Set metadata for the module """
        # Validation
        for key,val in dict(name=name, label=label).items():
            if val is not None:
                if not isinstance(val, str):
                    errormsg = f'Invalid value for {key}: must be str, not {type(val)}: {val}'
                    raise TypeError(errormsg)

        # Set values
        self.name = self._reconcile('name', name, self.__class__.__name__.lower())
        self.label = self._reconcile('label', label, self.name)
        return

    def define_pars(self, inherit=True, **kwargs): # TODO: think if inherit should default to true or false
        """ Create or merge Pars objects """
        if inherit: # Merge with existing
            self.pars.update(**kwargs, create=True)
        else: # Or overwrite
            self.pars = ss.Pars(**kwargs)
        return self.pars

    def update_pars(self, pars, **kwargs):
        """ Pull out recognized parameters, returning the rest """
        pars = sc.mergedicts(pars, kwargs)

        # Update matching module parameters
        matches = {}
        for key in list(pars.keys()): # Need to cast to list to avoid "dict changed during iteration"
            if key in self.pars:
                matches[key] = pars.pop(key)
        self.pars.update(matches)

        # Update module attributes
        metadata = {key:pars.get(key, self.pars.get(key)) for key in module_args}
        timepars = {key:pars.get(key, self.pars.get(key)) for key in ss.time.time_args}
        self.set_metadata(**metadata)
        self.t.update(**timepars)

        # Should be no remaining pars
        remaining = set(pars.keys()) - set(module_args + ss.time.time_args)
        if len(remaining):
            errormsg = f'{len(pars)} unrecognized arguments for {self.name}: {sc.strjoin(remaining)}'
            raise ValueError(errormsg)
        return

    def init_pre(self, sim, force=False):
        """
        Perform initialization steps

        This method is called once, as part of initializing a Sim. Note: after
        initialization, initialized=False until init_vals() is called (which is after
        distributions are initialized).
        """
        if force or not self.pre_initialized:
            self.sim = sim # Link back to the sim object
            ss.link_dists(self, sim, skip=ss.Sim) # Link the distributions to sim and module
            sim.pars[self.name] = self.pars
            sim.results[self.name] = self.results
            sim.people.add_module(self) # Connect the states to the people
            self.init_time() # Initialize the modules' time parameters and link them to the sim
            self.init_results()
            self.pre_initialized = True
        return

    def init_results(self):
        """ Initialize any results required; part of init_pre() """
        self.results.timevec = self.t.timevec # Store the timevec in the results for plotting
        return

    def init_post(self):
        """ Initialize the values of the states; the last step of initialization """
        for state in self.states:
            if not state.initialized:
                state.init_vals()
        self.initialized = True
        return

    def init_time(self, force=False):
        """ Initialize all time parameters by ensuring all parameters are initialized; part of init_post() """
        # Update time and initialize
        self.t.init(sim=self.sim) # Sets the absolute sim time vector

        # Find all time parameters in the module
        timepars = sc.search(self.pars, type=ss.TimePar) # Should it be self or self.pars?

        # Initialize them with the parent module
        for timepar in timepars.values():
            if force or not timepar.initialized:
                timepar.init(parent=self.t, die=False) # In some cases, the values can't be initialized; that's OK here
        return

    def match_time_inds(self, inds=None):
         """ Find the nearest matching sim time indices for the current module """
         if inds is None: inds = Ellipsis
         self_tvec = self.t.abstvec
         sim_tvec = self.sim.t.abstvec
         if len(self_tvec) == len(sim_tvec): # Shortcut to avoid doing matching
             return inds
         else:
             out = sc.findnearest(sim_tvec, [inds])
             return out

    def start_step(self):
        """ Tasks to perform at the beginning of the step """
        if self.dists is not None: # Will be None if no distributions are defined
            self.dists.jump_dt() # Advance random number generators forward for calls on this step
        return

    def step(self):
        """ Define how the module updates over time -- the key part of Starsim!! """
        errormsg = f'Module "{self.name}" does not define a "step" method: use "def step(self): pass" if this is intentional'
        raise NotImplementedError(errormsg)

    def finish_step(self):
        """ Define what should happen at the end of the step; at minimum, increment ti """
        self.t.ti += 1
        return

    def update_results(self):
        """ Perform any results updates on each timestep """
        pass

    def finalize(self):
        """ Perform any final operations, such as removing unneeded data """
        self.finalize_results()
        self.finalized = True
        return

    def finalize_results(self): # TODO: this is confusing, needs to be not redefined by the user, or called *after* a custom finalize_results()
        """ Finalize results """
        # Scale results
        for reskey, res in self.results.items():
            if isinstance(res, ss.Result) and res.scale:
                self.results[reskey] = self.results[reskey]*self.sim.pars.pop_scale
        return

    def define_states(self, *args, check=True):
        """
        Define states of the module with the same attribute name as the state

        Args:
            args (states): list of states to add
            check (bool): whether to check that the object being added is a state
        """
        for arg in args:
            if isinstance(arg, (list, tuple)):
                state = ss.State(*arg)
            elif isinstance(arg, dict):
                state = ss.State(**arg)
            else:
                state = arg

            if check:
                assert isinstance(state, ss.Arr), f'Could not add {state}: not an Arr object'

            # Add the state to the module
            setattr(self, state.name, state)
        return

    def define_results(self, *args, check=True):
        """ Add results to the module """
        for arg in args:
            if isinstance(arg, (list, tuple)):
                result = ss.Result(*arg)
            elif isinstance(arg, dict):
                result = ss.Result(**arg)
            else:
                result = arg

            # Update with module information
            result.update(module=self.name, shape=self.t.npts, timevec=self.t.timevec)

            # Add the result to the dict of results; does automatic checking
            self.results += result
        return

    @property
    def states(self):
        """
        Return a flat list of all states

        The base class returns all states that are contained in top-level attributes
        of the Module. If a Module stores states in a non-standard location (e.g.,
        within a list of states, or otherwise in some other nested structure - perhaps
        due to supporting features like multiple genotypes) then the Module should
        overload this attribute to ensure that all states appear in here.
        """
        return [x for x in self.__dict__.values() if isinstance(x, ss.Arr)] # TODO: use ndict

    @property
    def statesdict(self): # TODO: remove
        """
        Return a flat dictionary (objdict) of all states

        Note that name collisions may affect the output of this function
        """
        return sc.objdict({s.name:s for s in self.states})

    @classmethod
    def create(cls, name, *args, **kwargs):
        """
        Create a module instance by name

        Args:
            name (str): A string with the name of the module class in lower case, e.g. 'sir'
        """
        for subcls in ss.all_subclasses(cls):
            if subcls.__name__.lower() == name:
                return subcls(*args, **kwargs)
        else:
            raise KeyError(f'Module "{name}" did not match any known Starsim modules')

    @classmethod
    def from_func(cls, func):
        """ Create an module from a function """
        def step(mod): # TODO: see if this can be done more simply
            return mod.func(mod.sim)
        name = func.__name__
        mod = cls(name=name)
        mod.func = func
        mod.step = partial(step, mod)
        mod.step.__name__ = name # Manually add these in as for a regular class method
        mod.step.__self__ = mod
        return mod

    def to_json(self):
        """ Export to a JSON-compatible format """
        out = sc.objdict()
        out.type = self.__class__.__name__
        out.name = self.name
        out.label = self.label
        out.pars = self.pars.to_json()
        return out

    def plot(self):
        """ Plot all results in the module """
        with sc.options.with_style('fancy'):
            flat = sc.flattendict(self.results, sep=': ')
            timevec = self.t.timevec
            fig, axs = sc.getrowscols(len(flat), make=True)
            for ax, (k, v) in zip(axs.flatten(), flat.items()):
                ax.plot(timevec, v)
                ax.set_title(k)
                ax.set_xlabel('Year')
        return fig


class Analyzer(Module):
    """
    Base class for Analyzers. Analyzers are used to provide more detailed information
    about a simulation than is available by default -- for example, pulling states
    out of sim.people on a particular timestep before they get updated on the next step.

    The key method of the analyzer is ``step()``, which is called with the sim
    on each timestep.

    To retrieve a particular analyzer from a sim, use sim.get_analyzer().
    """
    pass


class Connector(Module):
    """
    Base class for Connectors, which mediate interactions between disease (or other) modules

    Because connectors can do anything, they have no specified structure: it is
    up to the user to define how they behave.
    """
    pass



"""
Compare two HIV simulations, one baseline and the other with ART
"""

# %% Imports and settings
import starsim as ss
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
import networkx as nx
import sys
import os
import argparse
import sciris as sc

default_n_agents = 25
# Three choices for network here, note that only the first two are common-random-number safe
network = ['stable_monogamy', 'EmbeddingNet'][1]

do_plot_graph = True
# Several choices for how to layout the graph when plotting
kind = ['radial', 'bipartite', 'spring', 'multipartite'][1]

do_plot_longitudinal = True
do_plot_timeseries = True

ss.options(multirng = True) # Can set multirng to False for comparison

figdir = os.path.join(os.getcwd(), 'figs', network)
sc.path(figdir).mkdir(parents=True, exist_ok=True)


class stable_monogamy(ss.SexualNetwork):
    """
    Very simple network for debugging in which edges are:
    1-2, 3-4, 5-6, ...
    """
    def __init__(self, **kwargs):
        # Call init for the base class, which sets all the keys
        super().__init__(**kwargs)
        return

    def init_pre(self, sim):
        n = len(sim.people._uid_map)
        n_edges = n//2
        self.edges.p1 = np.arange(0, 2*n_edges, 2) # EVEN
        self.edges.p2 = np.arange(1, 2*n_edges, 2) # ODD
        self.edges.beta = np.ones(n_edges)
        return


class Graph():
    def __init__(self, nodes, edges):
        self.graph = nx.from_pandas_edgelist(df=edges, source='p1', target='p2', edge_attr=True)
        self.graph.add_nodes_from(nodes.index)
        nx.set_node_attributes(self.graph, nodes.transpose().to_dict())
        return

    def draw_nodes(self, filter, pos, ax, **kwargs):
        inds = [i for i,n in self.graph.nodes.data() if filter(n)]
        nc = ['red' if nd['hiv'] else 'lightgray' for i, nd in self.graph.nodes.data() if i in inds]
        ec = ['green' if nd['on_art'] else 'black' for i, nd in self.graph.nodes.data() if i in inds]
        if inds:
            nx.draw_networkx_nodes(self.graph, nodelist=inds, pos=pos, ax=ax, node_color=nc, edgecolors=ec, **kwargs)
        return

    def plot(self, pos, edge_labels=False, ax=None):
        kwargs = dict(node_shape='x', node_size=250, linewidths=2, ax=ax, pos=pos)
        self.draw_nodes(lambda n: n['dead'], **kwargs)

        kwargs['node_shape'] = 'o'
        self.draw_nodes(lambda n: not n['dead'] and n['female'], **kwargs)
        
        kwargs['node_shape'] = 's'
        self.draw_nodes(lambda n: not n['dead'] and not n['female'], **kwargs)

        nx.draw_networkx_edges(self.graph, pos=pos, ax=ax)
        nx.draw_networkx_labels(self.graph, labels={i:int(a['cd4']) for i,a in self.graph.nodes.data()}, font_size=8, pos=pos, ax=ax)
        if edge_labels:
            nx.draw_networkx_edge_labels(self.graph, edge_labels={(i,j): int(a['dur']) if 'dur' in a else np.nan for i,j,a in self.graph.edges.data()}, font_size=8, pos=pos, ax=ax)
        return


class GraphAnalyzer(ss.Analyzer):
    ''' Simple analyzer to assess if common random numbers are working '''

    def __init__(self, **kwargs):
        super().__init__(**kwargs) # Initialize the Analyzer object

        self.graphs = {}
        return

    def init_pre(self, sim):
        self.initialized = True
        self.update_results(sim, init=True)
        return

    def update_results(self, sim, init=False):
        nodes = pd.DataFrame({
            'age': sim.people.age.values,
            'female': sim.people.female.values,
            'dead': sim.people.dead.values,
            'hiv': sim.people.hiv.infected.values,
            'on_art': sim.people.hiv.on_art.values,
            'cd4': sim.people.hiv.cd4.values,
        })

        edges = sim.people.networks[network.lower()].to_df()

        idx = sim.ti if not init else -1
        self.graphs[idx] = Graph(nodes, edges)
        return

    def finalize(self, sim):
        super().finalize(sim)
        return


def run_sim(n=25, rand_seed=0, intervention=False, analyze=False, lbl=None):
    ppl = ss.People(n)

    try:
        net_class = eval(network)
    except Exception as e:
        net_class = getattr(ss, network)
    ppl.networks = ss.ndict(net_class(), ss.MaternalNet())

    hiv_pars = {
        'beta': {network: [0.3, 0.25], 'maternal': [0.2, 0]},
        'init_prev': 0.25,
        'art_efficacy': 0.96,
    }
    hiv = ss.HIV(hiv_pars)

    art = ss.hiv.ART(0, 0.6) # 60% coverage from day 0

    pars = {
        'start': 1980,
        'end': 2020,
        'remove_dead': False, # So we can see who dies, sim results should not change with True
        'interventions': [art] if intervention else [],
        'rand_seed': rand_seed,
        'analyzers': [GraphAnalyzer()] if analyze else [],
    }
    sim = ss.Sim(people=ppl, diseases=[hiv], demographics=[ss.Pregnancy(), ss.Deaths()], pars=pars, label=lbl)
    sim.initialize()

    # Infect every other person, useful for exploration in conjunction with init_prev=0
    #sim.diseases['hiv'].set_prognoses(sim, np.arange(0,n,2), from_uids=None)

    sim.run()

    return sim


def run_scenario(n=10, rand_seed=0, analyze=True):
    sims = sc.parallelize(run_sim,
                          kwargs={'n':n, 'analyze': analyze, 'rand_seed': rand_seed},
                          iterkwargs=[{'intervention':False, 'lbl':'Baseline'}, {'intervention':True, 'lbl':'Intervention'}], die=True)

    for i, sim in enumerate(sims):
        sim.save(os.path.join(figdir, f'sim{i}.obj'))

    return sims


def getpos(ti, g1, g2, guess=None, kind='bipartite'):

    n1 = dict(g1[ti].graph.nodes.data())
    n2 = dict(g2[ti].graph.nodes.data())
    nodes = sc.mergedicts(n2, n1)
    n = len(nodes)

    if kind == 'radial':
        pos = {i:(np.cos(2*np.pi*i/n), np.sin(2*np.pi*i/n)) for i in range(n)}
        if guess:
            if len(guess) < n:
                pos = {i:(np.cos(2*np.pi*i/n), np.sin(2*np.pi*i/n)) for i in range(n)}

    elif kind == 'spring':
        pos = nx.spring_layout(g1[ti].graph, k=None, pos=guess, fixed=None, iterations=50, threshold=0.0001, weight=None, scale=1, center=None, dim=2, seed=None)
        if guess:
            pos = sc.mergedicts(pos, guess)

    elif kind == 'multipartite':
        pos = nx.multipartite_layout(g1[ti].graph, subset_key='female', align='vertical', scale=10, center=None)
        if guess:
            pos = sc.mergedicts(pos, guess)

        if guess:
            for i in guess.keys():
                pos[i] = (pos[i][0], guess[i][1]) # Keep new x but carry over y

    elif kind == 'bipartite':
        pos = {i:(nd['age'], 2*nd['female']-1 + np.random.uniform(-0.3, 0.3)) for i, nd in nodes.items()}

        if guess:
            for i in guess.keys():
                pos[i] = (pos[i][0], guess[i][1]) # Keep new x but carry over y

    return pos


def plot_graph(sim1, sim2):
    g1 = sim1.analyzers[0].graphs
    g2 = sim2.analyzers[0].graphs

    n = len(g1[-1].graph)
    el = n <= 25 # Draw edge labels

    fig, axv = plt.subplots(1, 2, figsize=(10,5))
    global ti
    timax = sim1.tivec[-1]

    global pos
    pos = {}
    pos[-1] = getpos(0, g1, g2, kind=kind)
    for ti in range(timax+1):
        pos[ti] = getpos(ti, g1, g2, guess=pos[ti-1], kind=kind)

    ti = -1 # Initial state is -1, representing the state before the first step

    def on_press(event):
        print('press', event.key)
        sys.stdout.flush()
        global ti, pos
        if event.key == 'right':
            ti = min(ti+1, timax)
        elif event.key == 'left':
            ti = max(ti-1, -1)

        # Clear
        axv[0].clear()
        axv[1].clear()

        g1[ti].plot(pos[ti], edge_labels=el, ax=axv[0])
        g2[ti].plot(pos[ti], edge_labels=el, ax=axv[1])
        fig.suptitle(f'Time is {ti} (use the arrow keys to change)')
        axv[0].set_title(sim1.label)
        axv[1].set_title(sim2.label)
        fig.canvas.draw()

    fig.canvas.mpl_connect('key_press_event', on_press)

    g1[ti].plot(pos[ti], edge_labels=el, ax=axv[0])
    g2[ti].plot(pos[ti], edge_labels=el, ax=axv[1])
    fig.suptitle(f'Time is {ti} (use the arrow keys to change)')
    axv[0].set_title(sim1.label)
    axv[1].set_title(sim2.label)
    return fig


def plot_ts():
    # Plot timeseries summary
    fig, axv = plt.subplots(2,1, sharex=True)
    axv[0].plot(sim1.tivec, sim1.results.hiv.n_infected, label=sim1.label)
    axv[0].plot(sim2.tivec, sim2.results.hiv.n_infected, ls=':', label=sim2.label)
    axv[0].set_title('HIV number of infections')

    axv[1].plot(sim1.tivec, sim1.results.hiv.new_deaths, label=sim1.label)
    axv[1].plot(sim2.tivec, sim2.results.hiv.new_deaths, ls=':', label=sim2.label)
    axv[1].set_title('HIV Deaths')

    plt.legend()
    return fig


def analyze_people(sim):
    p = sim.people
    ever_alive = ss.false(np.isnan(p.age))
    years_lived = np.full(len(p), sim.ti+1) # Actually +1 dt here, I think
    years_lived[p.dead] = p.ti_dead[p.dead]
    years_lived = years_lived[ever_alive] # Trim, could be more efficient
    age_initial = p.age[ever_alive].values - years_lived
    age_initial = age_initial.astype(np.float32) # For better hash comparability, there are small differences at float64

    df = pd.DataFrame({
        'id': [hash((p.slot[i], age_initial[i], p.female[i])) for i in ever_alive], # if slicing, don't need ._view,
        'age_initial': age_initial,
        'years_lived': years_lived,
        'ti_infected': p.hiv.ti_infected[ever_alive].values,
        'ti_art': p.hiv.ti_art[ever_alive].values,
        'ti_dead': p.ti_dead[ever_alive].values,

        # Useful for debugging, but not needed for plotting
        'slot': p.slot[ever_alive].values,
        'female': p.female[ever_alive].values,
    })
    df.replace(to_replace=ss.INT_NAN, value=np.nan, inplace=True)
    df['age_infected'] = df['age_initial'] + df['ti_infected']
    df['age_art']      = df['age_initial'] + df['ti_art']
    df['age_dead']     = df['age_initial'] + df['ti_dead']
    return df


def plot_longitudinal(sim1, sim2):

    df1 = analyze_people(sim1)
    df1['sim'] = 'Baseline'
    df2 = analyze_people(sim2)
    df2['sim'] = 'With ART'

    df = pd.concat([df1, df2]).set_index('id')

    df['ypos'] = pd.factorize(df.index.values)[0]
    N = df['sim'].nunique()
    height = 0.5/N

    fig, ax = plt.subplots(figsize=(10,6))

    # For the legend:
    plt.barh(y=0, left=0, width=1e-6, color='k', height=height, label='Alive')
    plt.barh(y=0, left=0, width=1e-6, color='m', height=height, label='Infected before birth')
    plt.barh(y=0, left=0, width=1e-6, color='r', height=height, label='Infected')
    plt.barh(y=0, left=0, width=1e-6, color='g', height=height, label='ART')
    plt.scatter(y=0, x=0, color='c', marker='|', label='Death')

    for n, (lbl, data) in enumerate(df.groupby('sim')):
        yp = data['ypos'] + n/(N+1) # Leave space

        ti_initial = np.maximum(-data['age_initial'], 0)
        ti_final = data['ti_dead'].fillna(40)
        plt.barh(y=yp, left=ti_initial, width=ti_final - ti_initial, color='k', height=height)

        # Infected before birth
        vertical = data['age_infected']<0
        plt.barh(y=yp[vertical], left=data.loc[vertical]['ti_infected'], width=ti_final[vertical]-data.loc[vertical]['ti_infected'], color='m', height=height)

        # Infected
        infected = ~data['ti_infected'].isna()
        ai = data.loc[infected]['age_infected'].values # Adjust for vertical transmission
        ai[~(ai<0)] = 0
        plt.barh(y=yp[infected], left=data.loc[infected]['ti_infected']-ai, width=ti_final[infected]-data.loc[infected]['ti_infected']+ai, color='r', height=height)

        # ART
        art = ~data['ti_art'].isna()
        plt.barh(y=yp[art], left=data.loc[art]['ti_art'], width=ti_final[art]-data.loc[art]['ti_art'], color='g', height=height)

        # Dead
        dead = ~data['ti_dead'].isna()
        plt.scatter(y=yp[dead], x=data.loc[dead]['ti_dead'], color='c', marker='|')

    ax.set_xlabel('Age (years)')
    ax.set_ylabel('UID')
    ax.legend()

    return fig


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-p', '--plot', help='Plot from a cached CSV file', type=str)
    parser.add_argument('-n', help='Number of agents', type=int, default=default_n_agents)
    parser.add_argument('-s', help='Rand seed', type=int, default=0)
    args = parser.parse_args()

    if args.plot:
        print('Reading files', args.plot)
        sim1 = sc.load(os.path.join(args.plot, 'sim1.obj'))
        sim2 = sc.load(os.path.join(args.plot, 'sim2.obj'))
    else:
        print('Running scenarios')
        [sim1, sim2] = run_scenario(n=args.n, rand_seed=args.s)

    if do_plot_longitudinal:
        plot_longitudinal(sim1, sim2)

    if do_plot_graph:
        plot_graph(sim1, sim2)

    if do_plot_timeseries:
        plot_ts()

    plt.show()
    print('Done')


"""
Vignette 03: "Hello world" results output
DECISIONS:
    - sim.plot() should do something sensible based on which modules are present
    - sim.save_csv() likewise
    - sim.plot() is equivalent to sim.plot('default')
    - sim.hiv.plot() is equivalent to sim.plot('hiv')
    - sim.plot_result('hiv.prevalence') is equivalent to plt.plot(sim.t, sim.results.hiv['prevalence'])
    - make a result_splitter as in RS.3.2 to get coinfection results

TO RESOLVE:
    - how can individual disease results be accessed?
    - what are the default plots for the sim? (i.e. from ss.Sim(modules='hpv').run().plot() )
    - what are the default plots for each module?
    - what style should we use for naming results?
    - what should come from sim.summary()?
"""

import stisim as ss

##################
# @RomeshA
##################
sim = ss.Sim(modules=ss.HPV())
sim.run()
sim.plot()  # Do something sensible based on which modules are present
plt.plot(sim.t, sim.results.hpv['new_infections'])
sim.save_csv()


##################
# @cliffckerr
##################
# CK.3.1 Plots for a single module
ss.Sim(modules='hpv').run().plot()

# CK.3.2 Plot for multiple modules
sim = ss.Sim(modules=['hiv', 'gonorrhea'])
sim.run()
sim.plot('default', 'hiv', 'gonorrhea')  # Plot default as well as disease-specific plots
print(sim.results.hiv) # All HIV results
print(sim.results['hiv']['incidence']) # A single result, can also be accessed as dict keys
print(sim.results.hiv_gonorrhea_coinfections) # Result populated by the HIV-gonorrhea connector
sim.plot() # Equivalent to sim.plot('default')
sim.hiv.plot() # Equivalent to sim.plot('hiv')
sim.gonorrhea.plot() # Equivalent to sim.plot('gonorrhea')


##################
# @robynstuart
##################
# RS.3.1 Plots for a single module
sim = ss.Sim(modules=ss.hiv())
sim.run()
ss.plot()  # Default plots: prevalence, deaths, new infections
sim.plot_result('hiv.prevalence')  # Alternative method

# RS.3.2 Plot for multiple modules
# By default, I don't think we will create results of the numbers of people with dual infection.
# However, we can make a result splitter to divide standard results by boolean people attributes.
results_by_hiv = ss.result_splitter(
    results=['syphilis.prev'],  # If blank, throw error
    by=['hiv.infected', 'hiv.susceptible', 'hiv.treated'],  # If blank, throw error?
    years=2020  # If blank, do all years
)
results_by_sex = ss.result_splitter(
    results=['syphilis.prev'],
    by=['female', 'male'],
)
results_by_pregnancy = ss.result_splitter(
    results=['syphilis.prev'],
    by=['pregnant'],
)
# Could maybe also allow people to pass in a function that returns a boolean array for people, e.g.
# NB I don't really like this. Another option could be:
#   class ppl(ss.People()):
#       @property
#       def young(self):
#           return self.age < 21
#   sim = cv.Sim(people=ppl())

youth = lambda sim: ss.true(sim.people.age < 21)
results_by_age = ss.result_splitter(
    results=['syphilis.new_infections'],
    by=[youth],  # or similar
)
sim.run(analyzers=[results_by_hiv, results_by_sex, results_by_pregnancy, results_by_age])

res = sim.get_analyzer().results
res.syphylis.prev.hiv.infected
res.syphylis.prev.hiv.susceptible
res.syphylis.prev.hiv.treated
res.syphylis.prev.youth  # TBC
